# coding: utf-8
"""
    RankInspection.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import matplotlib.ticker as ticker
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar, get_color
from OurPlotUtils import draw_as_image
from SvdDenoise import get_denoised_data

NUM_RANKS = 5

@ticker.FuncFormatter
def major_formatter(x, pos):
    return "%.0f" % x

class RankInspectionFigure( Dialog ):
    def __init__( self, parent, dialog, M, C, q, eslice, row):
        self.grab = 'local'     # used in grab_set
        self.parent = parent
        self.dialog = dialog
        self.M = M
        self.svd = np.linalg.svd(M)
        self.C = C
        self.q = q
        self.eslice = eslice
        self.row = row
        self.Cpinv = np.linalg.pinv(self.C)
        print(self.M.shape, self.C.shape, self.Cpinv.shape)
        Dialog.__init__( self, parent, "Rank Analysis", visible=False )

    def show( self ):
        self._show()

    def body( self, body_frame ):
        tk_set_icon_portable( self )

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        nrows = 5
        self.fig = fig = plt.figure( figsize=(15, 8) )
        gs = GridSpec( nrows*2, 2 )
        axLs = []
        for row in range(nrows):
            row2 = row*2
            ax = fig.add_subplot(gs[row2:row2+2, 0])
            axLs.append(ax)
        axR1 = fig.add_subplot(gs[0:nrows, 1])
        axR2 = fig.add_subplot(gs[nrows:nrows*2, 1])
        fig.tight_layout()

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        fig.suptitle("Rank Inspection in Elution Range %d-%d" % (self.eslice.start, self.eslice.stop), fontsize=20)

        try:
            ax0 = axLs[0]
            axins = inset_axes(ax0, width=2, height=1.4,
                                bbox_to_anchor=(0, 0, 2.5, 2.5),
                                bbox_transform=ax0.transAxes, loc='upper left')
            axins.set_axis_off()
            from_ax = self.dialog.axis_array[self.row][0]
            draw_as_image(axins, self.dialog.fig, from_ax)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)

        axR1.set_title("Top Five Singular Values", fontsize=16)
        axR1.set_ylabel("Scale", fontsize=12)
        axR1.xaxis.set_major_locator(plt.NullLocator())
        n = 5
        sigmas = self.svd[1]
        axR1.plot(np.arange(5), sigmas[0:n], ':', marker='o')

        axR2.set_title('Deviations caused by SVD-denoise with various ranks', fontsize=16)
        axR2.set_ylabel('nRMSD', fontsize=12)
        axR2.set_xlabel('Rank used in denoising', fontsize=12)

        axR2.xaxis.set_major_locator(ticker.MultipleLocator(1.00))
        axR2.xaxis.set_major_formatter(major_formatter)

        P0 = np.dot(self.M, self.Cpinv)
        y0 = P0[:,0]
        max_y = np.percentile(y0, 95)

        nrmsd_list = []
        for i in range(NUM_RANKS):
            rank = i + 1
            D = get_denoised_data(None, rank=rank, svd=self.svd)
            P_ = np.dot(D, self.Cpinv)
            y1 = P_[:,0]
            ax = axLs[i]
            ax.set_ylabel('$Log_{10}$(I)')
            ax.plot(self.q, np.log10(y0), label='A(q) from raw data')
            ax.plot(self.q, np.log10(y1), label='A(q) from denoised data')
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = xmin*0.6 + xmax*0.4
            ty = ymin*0.9 + ymax*0.1
            ax.text(tx, ty, "Rank %d denoise" % rank, fontsize=20, alpha=0.3)
            ax.legend()
            nrmsd = np.sqrt(np.average((y0 - y1)**2))/max_y
            nrmsd_list.append(nrmsd)

        nrmsd_array = np.array(nrmsd_list)
        ranks = np.arange(NUM_RANKS) + 1
        axR2.plot(ranks, nrmsd_array, marker='o')

        fig.subplots_adjust( top=0.88, bottom=0.08, left=0.07, right=0.95, wspace=0.2 )
        self.mpl_canvas.draw()

    def buttonbox( self ):
        bottom_frame = Tk.Frame(self)
        bottom_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.1

        tframe = Tk.Frame(bottom_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        box = Tk.Frame(bottom_frame)
        box.pack(side=Tk.RIGHT, padx=padx*2)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)



"""
    Solvers.SMC.SolverPyMC.py

    Copyright (c) 2024, SAXS Team, KEK-PF    
"""
import logging
from importlib import reload
import numpy as np
import pymc as pm
from molass_legacy.Optimizer.StateSequence import save_opt_params
from molass_legacy.Optimizer.OptimizerUtils import OptimizerResult

def dummy_objective(params):
    # print(params)
    return np.std(params)

def get_picklable_func(optimizer):
    return dummy_objective

class SolverPyMC:
    def __init__(self, optimizer):
        self.optimizer = optimizer
        self.cb_fh = optimizer.cb_fh
        self.nfev = 0
        self.logger = logging.getLogger(__name__)
    
    def minimize(self, objective, init_params, niter=100, seed=1234, bounds=None, callback=None):
        self.objective = objective

        lower = bounds[:,0]
        upper = bounds[:,1]

        with pm.Model() as no_grad_model:
            # uniform priors on params
            params = pm.Uniform(
                "params",
                shape=len(init_params),
                lower=lower,
                upper=upper,
                initval=init_params,
            )

            import Solvers.SMC.SolverDist
            reload(Solvers.SMC.SolverDist)
            from Solvers.SMC.SolverDist import create_custom_dist
            size = len(init_params)
            likelihood = create_custom_dist(objective, lower, upper, size, params)

        try:
            no_grad_model.compile_dlogp()
        except Exception as exc:
            print(type(exc))

        # import SMC.smc.sampling
        # reload(SMC.smc.sampling)
        # import SMC.smc.sampling as pmt

        with no_grad_model:
            # Use custom number of draws to replace the HMC based defaults
            idata_no_grad = pm.sample_smc(100)
            # idata_no_grad = pm.sample_smc(100, cores=1)
            # idata_no_grad = pm.sample_smc(100, kernel=pm.smc.MH)
            # idata_no_grad = pm.sample_smc(100, kernel=pm.smc.IMH)
            # idata_no_grad = pm.sample(100)
 
        return OptimizerResult(x=init_params, nit=niter, nfev=self.nfev)

    def callback(self, norm_params, f, accept):
        fv = self.objective(norm_params)
        self.logger.info("callback: fv=%.3g", fv)
        real_params = self.optimizer.to_real_params(norm_params)
        save_opt_params(self.cb_fh, real_params, fv, accept, self.nfev)
        return False



# coding: utf-8
"""
    MemAnnealing.py

    Mapped Elution Model using Simulated Annealing

    Copyright (c) 2020,2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline
from matplotlib.patches import Polygon
import molass_legacy.KekLib.DebugPlot as plt

USE_NUMBA = False
if USE_NUMBA:
    try:
        # for numba 1.49 or later
        from numba.core.decorators import jit
    except:
        from numba.decorators import jit

def gaussian(x, h, m, s):
    return h*np.exp(-((x-m)/s)**2)

class MemAnnealing:
    def __init__(self):
        pass

    def fit(self, y1, y2, K):

        x1 = np.arange(len(y1))
        x2 = np.arange(len(y2))
        spline1 = UnivariateSpline(x1, y1, s=0, ext=3)
        spline2 = UnivariateSpline(x2, y2, s=0, ext=3)

        if USE_NUMBA:

            @jit(nopython=True)
            def compute_ch2(x, y, mv, sv, spline, ax=None):
                if ax is not None:
                    ax.plot(x, y)

                y_ = np.zeros(len(x))
                chi2 = 0
                for k in range(K):
                    n = 2 + k*2
                    m = mv[k]
                    s = sv[k]
                    h = spline(m)
                    gy = gaussian(x, h, m, s)
                    y_ += gy
                    chi2 += np.sum((y_ - y)**2)
                return chi2

        else:

            def compute_ch2(x, y, mv, sv, spline, ax=None):
                if ax is not None:
                    ax.plot(x, y)

                y_ = np.zeros(len(x))
                chi2 = 0
                for k in range(K):
                    n = 2 + k*2
                    m = mv[k]
                    s = sv[k]
                    h = spline(m)
                    gy = gaussian(x, h, m, s)
                    y_ += gy
                    if ax is not None:
                        ax.plot(x, gy)
                        if k == K-1:
                            ax.plot(x, y_, ':')
                            poly_points = list(zip(x, y)) + list( reversed( list( zip(x, y_) ) ) )
                            diff_poly   = Polygon( poly_points, alpha=0.2 )
                            ax.add_patch(diff_poly)
                    chi2 += np.sum((y_ - y)**2)
                return chi2

        count = 0

        if USE_NUMBA:
            # spline does not seem to be supported.

            @jit(nopython=True)
            def obj_func(p):
                nonlocal count

                a, b = p[0:2]
                m = p[2:2+K]
                s = p[2+K:2+K+K]
                m_ = m*a + b
                s_ = s*a

                chi2 = compute_ch2(x1, y1, m_, s_, spline1) * compute_ch2(x2, y2, m, s, spline2)
                reverse_penality = 1 + max(0, (m[0] - m[1]))
                close_penality = max(1, (len(y2)/min(50, abs(m[0] - m[1])))**2)
                chi2_ = chi2 * reverse_penality*close_penality

                count += 1

                return chi2_
        else:

            def obj_func(p):
                nonlocal count

                a, b = p[0:2]
                m = p[2:2+K]
                s = p[2+K:2+K+K]
                m_ = m*a + b
                s_ = s*a

                # debug_plot = count%100 == 0
                debug_plot = False

                if debug_plot:
                    print([count], m)
                    plt.push()
                    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,7))
                    ax1, ax2 = axes
                    fig.suptitle(str([count]))
                else:
                    ax1, ax2 = None, None

                chi2 = compute_ch2(x1, y1, m_, s_, spline1, ax1) * compute_ch2(x2, y2, m, s, spline2, ax2)
                close_penality = max(1, (len(y2)/min(50, abs(m[0] - m[1])))**2)
                chi2_ = chi2 * close_penality

                if debug_plot:
                    print('reverse_penality=', reverse_penality)
                    print('close_penality=', close_penality)
                    print('chi2=', chi2)
                    print('chi2_=', chi2_)
                    fig.tight_layout()
                    plt.show()
                    plt.pop()

                count += 1

                return chi2_

        a_init = len(y1)/len(y2)
        b_init = 0
        m_init = [len(y2)*0.3, len(y2)*0.7]
        s_init = [50, 50]
        # range_init = [[a_init*0.3, a_init*3], [-100, 100], [180, 220], [250, 300], [10, 100], [10, 100]]
        range_init = [[a_init*0.8, a_init*1.2], [-200, 200], [180, 220], [230, 300], [10, 60], [10, 60]]
        x_range = np.array(range_init)
        p_init = np.average(x_range, axis=1)

        if True:
            from molass_legacy.KekLib.BasicUtils import Struct
            from SimulatedAnnealing import SimulatedAnnealing
            anneal = SimulatedAnnealing()
            x_min = x_range[:,0]
            x_max = x_range[:,1]
            def xconstaints(x):
                return x[2] < x[3]
            anneal.minimize(obj_func, xrange=x_range, start=p_init, seed=1234,
                        xconstaints=xconstaints, n=3000, m=1000)
            print('anneal.x=', anneal.x)
            result = Struct(x=anneal.x[-1])
        else:
            result = minimize(obj_func, p_init)
        self.spline1 = spline1
        self.spline2 = spline2
        return result



# coding: utf-8
"""
    Baseline.LpmInspect_for_paper.py

    Copyright (c) 2020-2021, SAXS Team, KEK-PF
"""
import copy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import animation
from matplotlib.patches import Polygon
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.DENSS.DenssUtils import fit_data
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy._MOLASS.SerialSettings import get_setting
from MatrixData import simple_plot_3d
from .Baseline import compute_baseline, better_integrative_curve, END_WIDTH
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from LPM import LPM_3d
from DataUtils import cut_upper_folders
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import BlinkingFrame

TITLE_FONTSIZE = 16
LEGEND_FONTSIZE = 11

class EcurveProxyCds:
    def __init__(self, ecurve, j_slice):
        self.ecurve = ecurve
        self.j_start = j_slice.start
        if self.j_start is None:
            self.j_start = 0
        self.y = ecurve.y[j_slice]

    def get_major_peak_info(self):
        ret_info = []
        for rec in self.ecurve.get_major_peak_info():
            ret_info.append([j - self.j_start for j in rec])
        return ret_info

def demo(root, in_folder, md, integral=False, for_all_q=True, peakno=None, whole_angle=True, smooth_line=True, show_preview=False):
    print(in_folder)

    if show_preview:
        from ThreeDimViewer import ThreeDimViewer
        dialog = ThreeDimViewer(root, md)
        dialog.show()

    fig = dplt.figure(figsize=(21, 11))
    demo_impl(fig, in_folder, md, integral=integral, for_all_q=for_all_q, peakno=peakno, whole_angle=whole_angle, smooth_line=smooth_line,
                show_scd=True,  # show_scd=False,
                show_lpm_fixed=True)
    dplt.show()

def demo_impl(fig, in_folder, md, integral=False, for_all_q=True, peakno=None, whole_angle=True,
                smooth_line=False, show_scd=False,
                show_lpm_fixed=False, suppress_titles=False):
    print('demo_impl: smooth_line=', smooth_line)

    xr = md.xr
    xr.set_elution_curve()
    if whole_angle:
        i_slice = slice(2, None)
    else:
        i_slice = xr.i_slice
    j_slice = xr.j_slice
    i_index = xr.e_index - i_slice.start

    whole_data = xr.data
    data = whole_data[i_slice,j_slice]
    error = xr.error.data[i_slice,j_slice]
    q = xr.vector[i_slice]
    if peakno is None:
        peakno = xr.e_curve.primary_peak_no

    if show_scd:
        from Conc.ConcDepend import ConcDepend
        # judge_holder, rdr_hints
        ecurve_for_cds = EcurveProxyCds(xr.e_curve, j_slice)
        cd = ConcDepend(q, data, error, ecurve_for_cds)
        cds_list = cd.compute_judge_info()
        cds0 = cds_list[peakno][1]

    gs = GridSpec(8,2)

    fig.suptitle("Baseline Correction Inspection for " + cut_upper_folders(in_folder), fontsize=28)

    # z = xr.e_curve.y[j_slice]
    z = data[i_index,:]
    start = 0 if j_slice.start is None else j_slice.start

    topx = xr.e_curve.peak_info[peakno][1] - start
    print('topx=', topx, start+topx)

    q1 = q[i_index]
    size = data.shape[1]
    x = np.ones(size)*q1
    y = np.arange(size)

    hw = 5
    yp = np.average(data[:,topx-hw:topx+hw], axis=1)
    ype = np.average(error[:,topx-hw:topx+hw], axis=1)

    diff_axes_info = []
    diff_ymin = None
    diff_ymax = None
    def update_diff_ylim(ax):
        ymin, ymax = ax.get_ylim()
        nonlocal diff_ymin, diff_ymax
        if diff_ymin is None or ymin < diff_ymin:
            diff_ymin = ymin
        if diff_ymax is None or ymax > diff_ymax:
            diff_ymax = ymax
    ylim_list = []
    axes_list = []

    end_slices = xr.e_curve.get_end_slices()

    def draw_scd_text(ax, cds):
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        tx = xmin*0.95 + xmax*0.05
        ty = np.power(10, np.log10(ymin)*0.9 + np.log10(ymax)*0.1)
        inc = cds - cds0
        sign  = '-' if inc < 0 else '+'
        ax.text(tx, ty, "CDS=%.2f %s %.2f" % (cds0, sign, abs(inc)), va="center", fontsize=30, alpha=0.5)

    cds_pairs = []
    for i, title in enumerate(["Linear Baseline", "Integral Baseline"]):
        gs_start = i*4
        gs_stop = (i+1)*4
        ax1 = fig.add_subplot(gs[gs_start:gs_stop,0])
        ax2 = fig.add_subplot(gs[gs_start:gs_stop,1])
        # ax2d = fig.add_subplot(gs[gs_stop-1:gs_stop,1])
        # ax3 = fig.add_subplot(gs[gs_start:gs_stop-1,2])
        # ax3d = fig.add_subplot(gs[gs_stop-1:gs_stop,2])
        # axes_list.append((ax2, ax3))

        for ax in [ax2]:
            ax.axes.get_xaxis().set_ticks([])

        # diff_axes_info.append((ax2, ax2d))

        ax1.set_title(title, fontsize=TITLE_FONTSIZE)
        ax1.plot(z, color='orange')

        integral = i == 1
        if i == 0:
            b = compute_baseline(z, integral=False, end_slices=end_slices)
        else:
            b, convex = better_integrative_curve(z, end_slices=end_slices)

        ax1.plot(b, ':', color='black', linewidth=3)
        # ax1.plot(topx, z[topx], 'o', color='blue')
        x = np.arange(len(z))
        # tslice = slice(topx-hw, topx+hw)
        # ax1.plot(x[tslice], z[tslice], color='blue')

        if i==1:
            for slice_ in end_slices:
                ax1.plot(y[slice_], z[slice_], 'o', color='yellow', markersize=3)

        if show_lpm_fixed:
            lpm0 = LPM_3d(data, integral=integral, for_all_q=True, e_index=i_index)
            yp_lpm0 = np.average(lpm0.data[:,topx-hw:topx+hw], axis=1)

        if i == 0:
            lpm1 = LPM_3d(data, ecurve_y=z, integral=integral, for_all_q=True, e_index=i_index)
            yp_lpm1 = np.average(lpm1.data[:,topx-hw:topx+hw], axis=1)

            ax2.set_yscale("log")
            ax2.set_xscale("log")

            title = "" if suppress_titles else "Correction Effects to the Scattering Curve (for all Q)"
            ax2.set_title(title, fontsize=TITLE_FONTSIZE)

            alpha = 0.3 if smooth_line else 1
            ax2.plot(q, yp, label='Average(top-%d:top+%d), No Correction' % (hw, hw),color='blue', alpha=alpha)
            if show_lpm_fixed:
                ax2.plot(q, yp_lpm0, label='Average(top-%d:top+%d), LPM (Fixed %%)' % (hw, hw), color='blue', alpha=alpha)
            ax2.plot(q, yp_lpm1, label='Average(top-%d:top+%d), LPM (Varying %%)' % (hw, hw), color='red', alpha=alpha)
            ylim_list.append(ax2.get_ylim())

            diff_y = (yp - yp_lpm1)/ype
            # ax2d.plot(q, diff_y, color='red')
            # update_diff_ylim(ax2d)

            if smooth_line:
                labels = ["No Correction", "Fixed %", "Varying %"]
                yp_fixed = yp_lpm0 if show_lpm_fixed else None
                for k, yp_ in enumerate([yp, yp_fixed, yp_lpm1]):
                    if k == 1 and not show_lpm_fixed:
                        continue
                    qc, ac, ec, dmax = fit_data(q, yp_, ype)
                    ax2.plot(qc, ac, color='C%d' % k, linewidth=3, label="DENSS fit-data (%s)" % labels[k])

            if show_scd:
                cd = ConcDepend(q, lpm1.data, error, ecurve_for_cds)
                cds_list = cd.compute_judge_info()
                cds1 = cds_list[peakno][1]

            def set_legend_ylim(ax):
                # ax.legend(fontsize=LEGEND_FONTSIZE)
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymax-4.5, ymax)

            set_legend_ylim(ax2)
        else:
            # ax2.axes.get_yaxis().set_ticks([])
            # ax2d.axes.get_xaxis().set_ticks([])
            # ax2d.axes.get_yaxis().set_ticks([])
            pass

        if i == 1:
            lpm2 = LPM_3d(data, ecurve_y=z, integral=integral, for_all_q=False, e_index=i_index, end_slices=end_slices)
            yp_lpm2 = np.average(lpm2.data[:,topx-hw:topx+hw], axis=1)

            ax2.set_yscale("log")
            ax2.set_xscale("log")

            title = "" if suppress_titles else "Correction Effects to the Scattering Curve (diff expanded)"
            ax2.set_title(title, fontsize=TITLE_FONTSIZE)

            labels = ["No Correction", "Diff Corrected"]
            colors = ["blue", "red"]

            ax2.plot(q, yp, label='Average(top-%d:top+%d), %s' % (hw, hw, labels[0]), color=colors[0], alpha=alpha)
            ax2.plot(q, yp_lpm2, label='Average(top-%d:top+%d), %s' % (hw, hw, labels[1]), color=colors[1], alpha=alpha)

            if smooth_line:
                for k, yp_ in enumerate([yp, yp_lpm2]):
                    qc, ac, ec, dmax = fit_data(q, yp_, ype)
                    ax2.plot(qc, ac, color=colors[k], linewidth=3, label="DENSS fit-data (%s)" % labels[k])

            if show_scd:
                cd = ConcDepend(q, lpm2.data, error, ecurve_for_cds)
                cds_list = cd.compute_judge_info()
                cds2 = cds_list[peakno][1]
                cds_pairs.append((cds1, cds2))

            set_legend_ylim(ax2)
            ylim_list.append(ax2.get_ylim())

            diff_y = (yp - yp_lpm2)/ype
            # ax2d.plot(q, diff_y, color='red')
            # update_diff_ylim(ax2d)

        else:
            pass
            # ax3.axes.get_yaxis().set_ticks([])
            # ax3d.axes.get_xaxis().set_ticks([])
            # ax3d.axes.get_yaxis().set_ticks([])

    for axes in diff_axes_info:
        for ax in axes[1:]:
            ax.set_ylim(diff_ymin, diff_ymax)

    ymin, ymax = np.average(np.array(ylim_list), axis=0)
    for axes, pair in zip(axes_list, cds_pairs):
        for ax, cds in zip(axes, pair):
            ax.set_ylim(ymin, ymax)
            draw_scd_text(ax, cds)

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)

    if False:

        for ax2, ax2d in diff_axes_info:
            ay0 = ax2.axes.get_position().y0
            for ax in [ax2d]:
                bbox = ax.axes.get_position()
                bbox.y1 = bbox.y0 + (ay0 - bbox.y0)*0.8
                ax.axes.set_position(bbox)

class LpmInspector(Dialog):
    def __init__(self, parent, md, suppress_titles=False):
        parent.config(cursor='wait')
        parent.update()
        self.parent = parent
        self.md = md
        self.suppress_titles = suppress_titles
        self.in_folder = get_setting('in_folder')
        Dialog.__init__(self, parent, "Baseline Inspector", visible=False)

    def show(self):
        self._show()

    def body(self, frame):
        cframe = Tk.Frame(frame)
        cframe.pack()
        bframe = Tk.Frame(frame)
        bframe.pack(fill=Tk.X)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)

        self.fig = plt.figure(figsize=(14, 11))
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()
        self.build_control_widgets(pframe)
        demo_impl(self.fig, self.in_folder, self.md,
                    smooth_line=self.smooth_lines.get(), show_lpm_fixed=False,
                    suppress_titles=self.suppress_titles)
        self.mpl_canvas.draw()

        self.parent.config(cursor='')

    def build_control_widgets(self, pframe):
        label = Tk.Label(pframe, text="Peak No: ")
        label.pack(side=Tk.LEFT)
        ecurve = self.md.xr.e_curve
        self.peakno = Tk.IntVar()
        self.peakno.set(ecurve.primary_peak_no + 1)
        max_peakno = len(ecurve.peak_info)
        self.spinbox = Tk.Spinbox(pframe, textvariable=self.peakno,
                                  from_=1, to=max_peakno, increment=1,
                                  justify=Tk.CENTER, width=6)
        self.spinbox.pack(side=Tk.LEFT)

        self.smooth_lines = Tk.IntVar()
        self.smooth_lines.set(0)
        toggle_cb = Tk.Checkbutton(pframe, text="show smooth lines", variable=self.smooth_lines)
        toggle_cb.pack(side=Tk.LEFT, padx=20, pady=5)

        self.show_scd = Tk.IntVar()
        self.show_scd.set(0)
        show_cds_cb = Tk.Checkbutton(pframe, text="show SCD", variable=self.show_scd)
        show_cds_cb.pack(side=Tk.LEFT, pady=5)

        self.redraw_btn_blink = BlinkingFrame(pframe)
        self.redraw_btn_blink.pack(side=Tk.LEFT, padx=20, pady=5)
        self.redraw_btn = Tk.Button(self.redraw_btn_blink, text="Redraw", width=10, command=self.redraw)
        self.redraw_btn.pack()
        self.redraw_btn_blink.objects = [self.redraw_btn]

        for v in [self.peakno, self.smooth_lines, self.show_scd]:
            v.trace("w", self.blink_start)

    def blink_start(self, *args):
        self.redraw_btn_blink.start()

    def redraw(self):
        self.redraw_btn_blink.stop()
        self.config(cursor='wait')
        self.update()
        self.fig.clf()
        peakno = self.peakno.get() - 1
        smooth_line=self.smooth_lines.get()
        show_scd = self.show_scd.get()
        demo_impl(self.fig, self.in_folder, self.md, peakno=peakno, smooth_line=smooth_line, show_scd=show_scd)
        self.mpl_canvas.draw()
        self.config(cursor='')
        


"""

    ElutionMapperAdjuster.py

    custom widget for mapping adjustment interface

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

"""

import logging
from molass_legacy.KekLib.OurTkinter import Tk, Font
from molass_legacy.KekLib.TkSupplements import BlinkingFrame
from molass_legacy.KekLib.ReadOnlyText import CopyableLabel
from CanvasFrame import CanvasFrame
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from .MappingParams import MappingParams

UV_BASELINE_METHOD_LPM  = 1
UV_LPM_OPTIONS = ['2D', '3D']
XR_BPA_OPTIONS = ['SCD 1', 'SCD 2']

class ElutionMapperAdjuster(Tk.Frame):
    def __init__( self, parent, canvas ):
        Tk.Frame.__init__( self, parent )

        self.parent     = parent
        self.canvas     = canvas
        self.mapper     = canvas.mapper
        self.logger     = logging.getLogger( __name__ )
        self.fixed_font = Font.Font( family="Courier", size=9 )
        self.doing_range_adj    = False
        self.use_xray_conc  = get_setting('use_xray_conc')
        self.use_mtd_conc  = get_setting('use_mtd_conc')
        self.enable_lrf_baseline = get_setting('enable_lrf_baseline')
        self.conc_type = self.mapper.get_conc_type()
        self.uv_lpm_menu = None
        self.xr_bpa_menu = None

        self.initial_std_diff   = canvas.initial_std_diff
        self.current_std_diff   = canvas.current_std_diff

        # --- Frame Settings ----------------------------------------------------
        self.i_frame = i_frame = Tk.Frame( self )
        i_frame.pack()

        iframe0 = Tk.Frame( i_frame )
        iframe0.pack( fill=Tk.X )

        iframe0.rowconfigure( 0, weight=1 )
        for j in range(3):
            iframe0.columnconfigure( j, weight=1 )

        iframe1_ = Tk.Frame( i_frame )
        iframe1_.pack( fill=Tk.X, pady=5 )
        for i in range(3):
            iframe1_.columnconfigure( i, weight=1 )

        self.iframe1 = iframe1 = Tk.Frame( iframe1_ )
        iframe1.grid( row=0, column=0 )
        iframe2 = Tk.Frame( iframe1_ )
        iframe2.grid( row=0, column=1, sticky=Tk.E )

        canvas_width = canvas.get_canvas_width()
        base_frame_width = ( canvas_width - 100 )//3

        m_frame0 = Tk.LabelFrame( iframe0, text="UV Absorbance Baseline Adjustment", labelanchor=Tk.N )
        m_frame0.grid( row=0, column=0, sticky=Tk.N + Tk.S + Tk.E + Tk.W, ipadx=5, ipady=10 )
        m_frame0_adjust = Tk.Frame( m_frame0, width=base_frame_width, height=1 )
        m_frame0_adjust.grid( row=0, column=0 )

        m_frame1 = Tk.LabelFrame( iframe0, text="Xray Scattering Baseline Adjustment", labelanchor=Tk.N )
        m_frame1.grid( row=0, column=1, sticky=Tk.N + Tk.S + Tk.E + Tk.W, padx=10, ipadx=5, ipady=10 )
        m_frame1_adjust = Tk.Frame( m_frame1, width=base_frame_width, height=1 )
        m_frame1_adjust.grid( row=0, column=0 )

        m_frame2 = Tk.LabelFrame( iframe0, text="Mapping Precision Measure", labelanchor=Tk.N )
        m_frame2.grid( row=0, column=2, sticky=Tk.N + Tk.S + Tk.E + Tk.W )
        m_frame2.columnconfigure(0, weight=1)
        m_frame2_adjust = Tk.Frame( m_frame2, width=base_frame_width, height=1 )
        m_frame2_adjust.grid( row=0, column=0, ipadx=5,  ipady=10 )

        self.blink_ready = False

        self.penalty_widgets = []
        self.allowance_vars = []
        self.trace_suppress = False

        # --- Absorbance Baseline Adjustment -----------------------------------
        if self.use_xray_conc or self.use_mtd_conc:
            # create a dummy varriable so that it can be manipulated in the same way as in the normal cases
            self.build_uv_absorbance_option_widgets_dummy()
        else:
            self.build_uv_absorbance_option_widgets( m_frame0 )

        # --- Scattering Baseline Adjustment -----------------------------------
        self.build_xray_scattering_option_widgets( m_frame1 )

        self.blink_ready = True

        # --- Mapping Precision Measure ----------------------------------------
        self.build_mapping_precision_measure_widgets( m_frame2_adjust )

        # --- Optimize Button --------------------------------------------------
        self.build_optimize_button( iframe1 )

        # --- Other Buttons ------------------------------------------------
        self.build_other_buttons( iframe2 )

        self.update_mapping_precision_measure( self.current_std_diff )

    def close_fig(self):
        self.cframe.close_fig()

    def get_conc_type( self ):
        return self.conc_type

    def start_blink( self ):
        if self.blink_ready:
            self.optimize_btn_blink.start()
            self.canvas.change_depedent_states( Tk.DISABLED )

    def stop_blink( self ):
        self.optimize_btn_blink.stop()
        self.canvas.change_depedent_states( Tk.NORMAL )

    def blink_start_tracer(self, *args):
        self.start_blink()

    def build_uv_absorbance_option_widgets( self, m_frame0 ):
        uv_baseline_frame = Tk.Frame( m_frame0 )
        uv_baseline_frame.grid( row=0, column=0, sticky=Tk.W )

        self.uv_baseline_opt = Tk.IntVar()
        self.uv_baseline_opt.set( get_setting( 'uv_baseline_opt' ) )
        self.uv_baseline_type = Tk.IntVar()
        self.uv_baseline_type.set( get_setting( 'uv_baseline_type' ) )

        lrf_base_label = "LRF" if self.enable_lrf_baseline else "shited"
        radio_button_params = [(1, "linear"), (4, lrf_base_label), (5, "integral")]

        self.uv_radio_buttons = []
        self.uv_baseline_opt_dependents = []
        grid_row = -1
        for i, t in enumerate( [ "No correction to input data", "Correction to input data" ] ):
            grid_row += 1
            state = Tk.NORMAL
            b = Tk.Radiobutton( uv_baseline_frame, text=t, variable=self.uv_baseline_opt, value=i, state=state )
            b.grid( row=grid_row, column=0, sticky=Tk.W )
            self.uv_radio_buttons.append( b )
            if i == 1:
                grid_row += 1
                frame = Tk.Frame( uv_baseline_frame )
                frame.grid( row=grid_row, column=0, sticky=Tk.W )
                space = Tk.Frame( frame, width=40 )
                space.grid( row=0, column=0 )
                label = Tk.Label( frame, text="using LPM {" )
                label.grid( row=0, column=1 )
                self.uv_baseline_opt_dependents.append( label )
                state_ = self.get_uv_base_const_state()
                for j, t in radio_button_params:
                    b = Tk.Radiobutton( frame, text=t, variable=self.uv_baseline_type, value=j, state=state_ )
                    b.grid( row=0, column=j+2 )
                    self.uv_baseline_opt_dependents.append( b )

                label = Tk.Label( frame, text="} baseline" )
                label.grid( row=0, column=j+3 )
                self.uv_baseline_with_bpa = Tk.IntVar()
                self.uv_baseline_with_bpa.set(get_setting('uv_baseline_with_bpa'))
                cb = Tk.Checkbutton( frame, text="with BPA", variable=self.uv_baseline_with_bpa, state=state )
                cb.grid( row=0, column=j+4, sticky=Tk.W )
                self.uv_baseline_with_bpa_cb = cb
                self.uv_baseline_with_bpa.trace("w", self.blink_start_tracer)
                self.uv_baseline_opt_dependents.append( cb )

        grid_row += 1
        frame = Tk.Frame( uv_baseline_frame )
        frame.grid( row=grid_row, column=0, sticky=Tk.W )

        self.uv_baseline_adjust = Tk.IntVar()
        self.uv_baseline_adjust.set( get_setting( 'uv_baseline_adjust' )  )
        cb = Tk.Checkbutton( frame, text="Linear adjutment by iterated optimization",
                                variable=self.uv_baseline_adjust )
        cb.grid( row=0, column=0, sticky=Tk.W )
        self.uv_baseline_adjust_cb = cb

        self.uv_deviation_ratio = Tk.DoubleVar()
        self.uv_deviation_ratio.set( get_setting( 'dev_allow_ratio' ) )
        self.build_penalty_widgets( frame, self.uv_deviation_ratio )

        self.uv_baseline_opt.trace( 'w', self.uv_baseline_opt_tracer )
        self.uv_baseline_type.trace( 'w', self.blink_start_tracer )
        self.uv_baseline_adjust.trace( 'w', self.baseline_adjust_tracer )
        self.uv_deviation_ratio.trace( 'w', lambda *args: self.deviation_ratio_tracer( 0 ) )

    def build_uv_absorbance_option_widgets_dummy(self):
        self.uv_deviation_ratio = Tk.DoubleVar()
        self.allowance_vars.append(self.uv_deviation_ratio)

    def lpm_right_button(self, event):
        print('lpm_right_button', event)

        if self.uv_lpm_menu is None:
            self.uv_lpm_option = Tk.IntVar()
            self.uv_lpm_option.set(get_setting('uv_lpm_option'))
            self.uv_lpm_menu = Tk.Menu( self, tearoff=0 )
            self.uv_lpm_menu.add_command( label=UV_LPM_OPTIONS[0], command=lambda: self.set_uv_lpm_option(0) )
            self.uv_lpm_menu.add_command( label=UV_LPM_OPTIONS[1], command=lambda: self.set_uv_lpm_option(1) )

        defaultbg = self.uv_lpm_menu.cget('bg')
        uv_lpm_option = self.uv_lpm_option.get()
        for k in range(2):
            color = 'cyan' if uv_lpm_option == k else defaultbg
            self.uv_lpm_menu.entryconfigure(k, background=color)
        self.uv_lpm_menu.post(event.x_root, event.y_root)

    def set_uv_lpm_option(self, opt):
        self.uv_lpm_option.set(opt)
        set_setting('uv_lpm_option', opt)

    def get_uv_base_const_state( self ):
        return Tk.NORMAL if self.uv_baseline_opt.get() == 1 else Tk.DISABLED

    def uv_baseline_opt_tracer( self, *args ):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox

        uv_adjust_suppressed = get_setting('uv_adjust_suppressed')
        if not uv_adjust_suppressed:
            if self.uv_baseline_opt.get() == 0:
                # when no correction, choosing adjust is unusual
                self.uv_baseline_adjust.set(0)
                MessageBox.showwarning('Linear adjustment change',
                    'Be aware that choosing "No correction" forces\n'
                    '"Linear adjustment" to be "OFF" by default.\n'
                    'You should change it manually if not appropriate.',
                    parent=self
                    )
            else:
                self.uv_baseline_adjust.set(1)
                MessageBox.showwarning('Linear adjustment change',
                    'Be aware that choosing "Correction" forces\n'
                    '"Linear adjustment" to be "ON" by default.\n'
                    'You should change it manually if not appropriate.',
                    parent=self
                    )

        state =self.get_uv_base_const_state()
        for w in self.uv_baseline_opt_dependents:
            w.config( state=state )
        self.start_blink()

    def build_xray_scattering_option_widgets( self, m_frame1 ):
        xray_baseline_frame = Tk.Frame( m_frame1 )
        xray_baseline_frame.grid( row=0, column=0, sticky=Tk.W )

        self.xray_baseline_opt = Tk.IntVar()
        self.xray_baseline_opt.set( get_setting( 'xray_baseline_opt' ) )
        self.xray_baseline_type = Tk.IntVar()
        self.xray_baseline_type.set( get_setting( 'xray_baseline_type' ) )
        self.xray_radio_buttons = []
        self.xray_base_const_dependents = []

        grid_row = -1
        indent = 40

        for i, t in enumerate( [ "No correction to input data", "Correction to input data with method option:" ] ):
            grid_row += 1
            state = Tk.DISABLED if self.use_mtd_conc else Tk.NORMAL
            b = Tk.Radiobutton( xray_baseline_frame, text=t, variable=self.xray_baseline_opt, value=i, state=state )
            b.grid( row=grid_row, column=0, sticky=Tk.W )
            self.xray_radio_buttons.append( b )
            if i == 1:
                grid_row += 1
                frame = Tk.Frame( xray_baseline_frame )
                frame.grid( row=grid_row, column=0 , sticky=Tk.W)
                space = Tk.Frame( frame, width=indent )
                space.grid( row=0, column=0 )
                state_ = self.get_xray_base_const_state()
                label = Tk.Label( frame, text="using LPM {", state=state_ )
                label.grid( row=0, column=1 )
                self.xray_base_const_dependents.append( label )
                value_list = [1, 5]
                text_list = ["linear", "integral"]

                for j, t in zip(value_list, text_list):
                    # state__ = state_ if j < 3 else Tk.DISABLED
                    state__ = state_ 
                    b = Tk.Radiobutton( frame, text=t, variable=self.xray_baseline_type, value=j, state=state__ )
                    b.grid( row=0, column=j+2 )
                    self.xray_base_const_dependents.append( b )
                label = Tk.Label( frame, text="} baseline", state=state_ )
                label.grid( row=0, column=j+3 )
                self.xray_base_const_dependents.append( label )

                self.xray_baseline_with_bpa = Tk.IntVar()
                self.xray_baseline_with_bpa.set(get_setting('xray_baseline_with_bpa'))
                cb = Tk.Checkbutton( frame, text="with BPA", variable=self.xray_baseline_with_bpa, state=state )
                cb.grid( row=0, column=j+4, sticky=Tk.W )
                cb.bind("<Button-3>", self.bpa_right_button)
                self.xray_baseline_with_bpa_cb = cb
                self.xray_baseline_with_bpa.trace("w", self.blink_start_tracer)
                self.xray_base_const_dependents.append( cb )

        grid_row += 1
        frame = Tk.Frame( xray_baseline_frame )
        frame.grid( row=grid_row, column=0, sticky=Tk.W )

        self.xray_baseline_adjust = Tk.IntVar()
        xray_baseline_adjust = get_setting('xray_baseline_adjust')
        disable_xray_adjust = get_setting('disable_xray_adjust')
        if disable_xray_adjust:
            xray_baseline_adjust = 0
        self.xray_baseline_adjust.set(xray_baseline_adjust)

        if disable_xray_adjust:
            dev_allow_ratio = 1
        else:
            dev_allow_ratio = get_setting('dev_allow_ratio')
            state = Tk.DISABLED if self.use_xray_conc or self.use_mtd_conc else Tk.NORMAL
            cb = Tk.Checkbutton( frame, text="Linear adjutment by iterated optimization",
                                    variable=self.xray_baseline_adjust, state=state )
            cb.grid( row=0, column=0, sticky=Tk.W )

        self.xray_deviation_ratio = Tk.DoubleVar()
        self.xray_deviation_ratio.set(1 - dev_allow_ratio)
        self.build_penalty_widgets( frame, self.xray_deviation_ratio )

        self.xray_baseline_opt.trace( 'w', self.xray_baseline_opt_tracer )
        self.xray_baseline_type.trace( 'w', self.blink_start_tracer )
        self.xray_baseline_adjust.trace( 'w', self.baseline_adjust_tracer )
        self.xray_deviation_ratio.trace( 'w', lambda *args: self.deviation_ratio_tracer( 1 ) )
        self.baseline_adjust_tracer()

    def build_penalty_widgets( self, frame, var ):
        self.allowance_vars.append( var )

        label = Tk.Label( frame, text="with deviation ratio" )
        label.grid( row=0, column=1 )
        # label.grid_forget()
        self.penalty_widgets.append( label )
        pw_entry = Tk.Spinbox( frame, textvariable=var,
                                        from_=0, to=1, increment=0.1, 
                                        justify=Tk.CENTER, width=6 )
        pw_entry.grid( row=0, column=2, padx=2 )
        # pw_entry.grid_forget()
        self.penalty_widgets.append( pw_entry )

    def deviation_ratio_tracer( self, i ):
        if self.trace_suppress:
            return

        val = self.allowance_vars[i].get()
        self.allowance_vars[1-i].set( float('%g' % ( 1 - val ) ) )
        if val == 0:
            if i == 0:
                self.uv_baseline_adjust.set( 0 )
            else:
                self.xray_baseline_adjust.set( 0 )

        self.start_blink()

    def bpa_right_button(self, event):
        print('bpa_right_button', event)

        if self.xr_bpa_menu is None:
            self.xr_bpa_option = Tk.IntVar()
            self.xr_bpa_option.set(get_setting('xr_bpa_option'))
            self.xr_bpa_menu = Tk.Menu( self, tearoff=0 )
            self.xr_bpa_menu.add_command( label=XR_BPA_OPTIONS[0], command=lambda: self.set_xr_bpa_option(1) )
            self.xr_bpa_menu.add_command( label=XR_BPA_OPTIONS[1], command=lambda: self.set_xr_bpa_option(2) )

        defaultbg = self.xr_bpa_menu.cget('bg')
        xr_bpa_option = self.xr_bpa_option.get()
        for k in range(2):
            color = 'cyan' if xr_bpa_option == k+1 else defaultbg
            self.xr_bpa_menu.entryconfigure(k, background=color)
        self.xr_bpa_menu.post(event.x_root, event.y_root)

    def set_xr_bpa_option(self, cd):
        print('bpa cd=', cd)
        self.xr_bpa_option.set(cd)
        set_setting('xr_bpa_option', cd)

    def get_xray_base_const_state( self ):
        return Tk.NORMAL if self.xray_baseline_opt.get() == 1 else Tk.DISABLED

    def xray_baseline_opt_tracer( self, *args ):
        if self.xray_baseline_opt.get() == 0:
            # when no correction, choosing adjust is unusual
            self.xray_baseline_adjust.set(0)
        else:
            if self.xray_baseline_type.get() == 0:
                self.xray_baseline_type.set(1)

        state =self.get_xray_base_const_state()
        for w in self.xray_base_const_dependents:
            w.config( state=state )
        self.start_blink()

    def get_xray_base_degree( self ):
        return self.xray_baseline_type.get() + 1

    def baseline_adjust_tracer( self, *args ):
        if self.use_xray_conc or self.use_mtd_conc:
            uv_baseline_adjust  = self.xray_baseline_adjust.get()
        else:
            uv_baseline_adjust  = self.uv_baseline_adjust.get()
        xray_baseline_adjust    = self.xray_baseline_adjust.get()
        self.start_blink()
        if uv_baseline_adjust == 1 and xray_baseline_adjust == 1:
            activate = True
        else:
            activate = False
        for i, w in enumerate( self.penalty_widgets ):
            j = i % 2
            if activate:
                w.grid( row=0, column=2+j, padx=j*2 )
            else:
                w.grid_forget()

        allow_ratio = self.allowance_vars[0].get()
        if 0 < allow_ratio and allow_ratio < 1:
            pass
        else:
            if allow_ratio == 0:
                allow_ratio = 0.1
            else:
                allow_ratio = 0.9
            self.trace_suppress = True
            self.allowance_vars[0].set( allow_ratio )
            self.allowance_vars[1].set( float( '%g' % (1 - allow_ratio) ) )
            self.trace_suppress = False

    def build_mapping_precision_measure_widgets( self, m_frame2_adjust ):
        self.chi_labels = []
        label1 = CopyableLabel( m_frame2_adjust, text="", font=self.fixed_font )
        label1.grid( row=0, column=0, padx=5 )
        self.chi_labels.append( label1 )
        label2 = CopyableLabel( m_frame2_adjust, text="", font=self.fixed_font )
        label2.grid( row=1, column=0, padx=5 )
        self.chi_labels.append( label2 )

        self.cframe = CanvasFrame( m_frame2_adjust, figsize=( 3.0, 0.7 ) )
        self.cframe.grid( row=0, column=1, rowspan=2 )
        self.ax_for_chi = None

        self.sync_options = Tk.IntVar()
        feature_mapped = self.mapper.feature_mapped
        mapper_sync_options = get_setting('mapper_sync_options')
        if mapper_sync_options is None:
            mapper_sync_options = 1 if feature_mapped else 0
        self.sync_options.set(mapper_sync_options)
        sync_opt_frame = Tk.Frame(m_frame2_adjust)
        sync_opt_frame.grid( row=2, column=0, columnspan=2, sticky=Tk.W + Tk.E )
        sync_opt_frame.columnconfigure( 0, weight=1 )
        sync_opt_frame.columnconfigure( 1, weight=1 )
        sync_opt_label = Tk.Label(sync_opt_frame, text="Syncronization Options")
        sync_opt_label.grid(row=0, column=0, sticky=Tk.W, padx=5)
        sybc_opt_btn_frame = Tk.Frame(sync_opt_frame)
        sybc_opt_btn_frame.grid(row=1, column=0, padx=35)
        manual_time_scale = get_setting('manual_time_scale')
        self.sync_opt_btns = []
        # for j, t in enumerate( [ "rmsd finally", "features only", "manual adjust" ] ):
        for j, t in [(0, "automatic"), (2, "manual adjust")]:
            if j == 2 and manual_time_scale is None:
                state = Tk.DISABLED
            else:
                state = Tk.NORMAL
            b = Tk.Radiobutton( sybc_opt_btn_frame, text=t, variable=self.sync_options, value=j, state=state )
            b.grid( row=0, column=j, sticky=Tk.W, padx=5 )
            self.sync_opt_btns.append(b)

        self.sync_options.trace('w', self.sync_options_tracer)

    def sync_options_tracer(self, *args):
        self.start_blink()

    def update_mapping_precision_measure( self, std_diff=None ):
        self.current_std_diff = std_diff
        chi_list = [ self.initial_std_diff ]
        std_diff_ = self.initial_std_diff if std_diff is None else std_diff
        chi_list.append( std_diff_ )

        for i, chi in enumerate( chi_list ):
            varname = 'Initial' if i==0 else 'Current'
            self.chi_labels[i].config( text='%s nRMSD=%.5g' % ( varname, chi ) )

        def draw_std_diff_func( fig ):
            if self.ax_for_chi is None:
                self.ax_for_chi = fig.add_subplot( 111 )
            else:
                self.ax_for_chi.cla()
            self.ax_for_chi.set_axis_off()
            self.ax_for_chi.set_xlim( 0, 1 )
            self.ax_for_chi.set_ylim( 0, 1 )
            prev_chi = None
            for i, chi in enumerate( chi_list ):
                y = 0.75 - 0.5*i
                if chi <= 0.3:
                    color = 'green'
                elif chi <= 0.5:
                    color = 'orange'
                else:
                    color = 'red'
                if prev_chi is not None and chi > prev_chi:
                    color = 'red'
                self.ax_for_chi.plot( [ 0, chi ], [ y,  y ], color=color, linewidth=10, solid_capstyle='butt', alpha=0.5 )
                prev_chi = chi

        self.cframe.draw( draw_std_diff_func )
        self.canvas.update_guide_message()

    def build_optimize_button( self, iframe1 ):
        label11 = Tk.Label( iframe1, text="Press " )
        label11.pack( side=Tk.LEFT )

        self.optimize_btn_blink = BlinkingFrame( iframe1 )
        self.optimize_btn_blink.pack( side=Tk.LEFT )

        opt_btn_state = Tk.DISABLED if self.canvas is None or self.use_mtd_conc else Tk.NORMAL
        self.optimize_btn = Tk.Button( self.optimize_btn_blink, text="Optimize", command=self.optimize, state=opt_btn_state )
        self.optimize_btn.pack()

        self.optimize_btn_blink.objects = [self.optimize_btn]

        label12 = Tk.Label( iframe1, text=' to get adjusted with the specified parameters above; Press "OK" button below when it is appropriate.' )
        label12.pack( side=Tk.LEFT )

    def build_other_buttons( self, iframe2 ):
        self.manual_sync_btn = Tk.Button( iframe2, text="Manual Adjust", command=self.show_manual_adjuster )
        self.manual_sync_btn.pack( side=Tk.RIGHT, anchor=Tk.E, padx=5 )
        self.enable_mapping_anim    = get_setting( 'enable_mapping_anim' ) == 1
        if self.enable_mapping_anim:
            self.anim_btn = Tk.Button( iframe2, text="Animation", command=self.show_animation )
            self.anim_btn.pack( side=Tk.RIGHT, anchor=Tk.E, padx=5 )

    def optimize( self, helper_info=None ):
        self.change_cursor( 'wait' )

        if helper_info is None:
            helper_info = self.canvas.get_helper_info()

        if not self.use_xray_conc:
            pre_recog = self.canvas.serial_data.pre_recog
            btype = self.uv_baseline_type.get()
            self.mapper.absorbance.compute_base_curve(pre_recog, btype)

        while True:
            try:
                self.optimize_impl( helper_info )
                break
            except RuntimeError as exc:
                from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
                exception = exc
                etb = ExceptionTracebacker()
                self.logger.warning('optimize_impl failed etb=%s', str(etb))
                assert False

        self.change_cursor( '' )

    def change_cursor( self, cursor ):
        for w in self.canvas.get_cursor_widgets():
            w.config( cursor=cursor )   # does not work
            w.update()

    def update_proportional_uv_curve(self):
        self.stop_blink()
        self.mapper.x_base = self.mapper.compute_xray_baseline( self.get_mapping_params() )
        self.mapper.a_base = self.mapper.x_base / self.mapper.x_curve.max_y
        self.mapper.A_init = 1
        self.mapper.B_init = 0
        self.canvas.draw( clear=True )

    def get_mapping_params( self ):
        if self.use_xray_conc or self.use_mtd_conc:
            uv_baseline_opt = self.xray_baseline_opt.get()
            if self.use_xray_conc:
                uv_baseline_type = 0
            else:
                uv_baseline_type = self.xray_baseline_type.get()
            uv_baseline_adjust = self.xray_baseline_adjust.get()
            uv_baseline_with_bpa = self.xray_baseline_with_bpa.get()
        else:
            uv_baseline_opt = self.uv_baseline_opt.get()
            uv_baseline_type = self.uv_baseline_type.get()
            uv_baseline_adjust = self.uv_baseline_adjust.get()
            uv_baseline_with_bpa = self.uv_baseline_with_bpa.get()
        return MappingParams([
            ( 'uv_baseline_opt',            uv_baseline_opt ),
            ( 'uv_baseline_type',           uv_baseline_type ),
            ( 'uv_baseline_adjust',         uv_baseline_adjust ),
            ( 'uv_baseline_with_bpa',       uv_baseline_with_bpa ),
            ( 'xray_baseline_opt',          self.xray_baseline_opt.get() ),
            ( 'xray_baseline_type',         self.xray_baseline_type.get() ),
            ( 'xray_baseline_adjust',       self.xray_baseline_adjust.get() ),
            ( 'xray_baseline_with_bpa',     self.xray_baseline_with_bpa.get() ),
            ( 'dev_allow_ratio',            self.allowance_vars[0].get() ),
            ])

    def optimize_impl( self, helper_info ):
        self.change_cursor( 'wait' )

        self.stop_blink()

        params = self.get_mapping_params()
        sync_options = self.sync_options.get()
        self.mapper.optimize( opt_params=params, sync_options=sync_options, helper_info=helper_info )

        self.canvas.update_helper_info( helper_info )
        self.canvas.draw( clear=True )
        self.canvas.update_button_colors()

        # Pending:
        # self.canvas.update_range_info( update_list=True )   # update range list since range info may have been changed.
        self.current_std_diff   = self.mapper.std_diff
        self.update_mapping_precision_measure( self.current_std_diff )

        self.change_cursor( '' )

    def show_manual_adjuster(self, debug=True):
        if debug:
            from importlib import reload
            import Mapping.ManualAdjuster
            reload(Mapping.ManualAdjuster)
        from .ManualAdjuster import ManualAdjuster
        ms = ManualAdjuster(self.canvas, self.canvas.serial_data, self.mapper)
        ms.show()
        if ms.applied:
            self.sync_opt_btns[1].config(state=Tk.NORMAL)
            self.sync_options.set(2)

    def show_animation( self ):
        self.change_cursor( 'wait' )
        from molass_legacy.Mapping.ElutionMapperAnimator  import ElutionMapperAnimation
        anim = ElutionMapperAnimation( self.parent, self.canvas.serial_data )
        self.change_cursor( '' )
        anim.show()



"""
    GuinierTools.RgCurveUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.NumericalUtils import safe_ratios
# from molass_legacy.Trimming.Sigmoid import sigmoid

VALID_QUIALTY_LIMIT = 0.01
VALID_BASE_QUALITY = 0.3

def convert_to_milder_qualities(qualities):
    # task: this is a temporary fix. deeper improvement is required. 
    ret_qualities = qualities.copy()
    valid = qualities > VALID_QUIALTY_LIMIT
    ret_qualities[valid] = VALID_BASE_QUALITY + (1 - VALID_BASE_QUALITY)*qualities[valid]
    return ret_qualities

def get_connected_curve_info(rg_curve, debug=False):
    segments = rg_curve.get_curve_segments()
    qualities = np.concatenate(rg_curve.qualities)      # basic qualiteis; see RgProcess.RgCurve.py

    x_list = []
    y_list = []
    rg_list = []
    valid_bool_all = np.zeros(len(rg_curve.x), dtype=bool)
    valid_bool_seg_list = []
    k = 0
    for state, slice_ in zip(rg_curve.states, rg_curve.slices):
        if state == 0:           
            continue

        x_, y_, rg_ = segments[k]
        x_list.append(x_)
        y_list.append(y_)
        rg_list.append(rg_)
        bvec = rg_curve.qualities[k] > VALID_QUIALTY_LIMIT
        valid_bool_all[slice_] = bvec
        valid_bool_seg_list.append(bvec)
        k += 1

    x_ = np.concatenate(x_list)
    y_ = np.concatenate(y_list)
    rgv = np.concatenate(rg_list)
    valid_bool_seg = np.concatenate(valid_bool_seg_list)

    if debug:
        print(len(x_), len(y_), len(rgv), len(qualities))
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("get_connected_curve_info: debug")
            axt = ax.twinx()
            axt.grid(False)
            for dx, dy, drg in zip(x_list, y_list, rg_list):
                ax.plot(dx, dy)
                axt.plot(dx, drg, color='gray')
            # axt.plot(x_[valid_bools], rgv[valid_bools], 'o', color="green", alpha=0.3, label="valid $R_g's$")
            axt.plot(x_, qualities*100, color='green', alpha=0.5)
            fig.tight_layout()
            plt.show()

    return x_, y_, rgv, qualities, (valid_bool_all, valid_bool_seg)

def get_reconstructed_curve(size, valid_bools, Cxr, rg_params):
    valid_bool_all, valid_bool_seg = valid_bools
    ty_ = np.sum(Cxr, axis=0)[valid_bool_all]
    rrgv = np.zeros(size)
    for cy, rg in zip(Cxr[:-1], rg_params):
        rrgv += cy[valid_bool_all]/ty_*rg
    return rrgv

def compute_rg_curves(x, xr_weights, rg_params, xr_cy_list, xr_ty, rg_curve, debug=False):
    # recompute without mask for plot

    t_rg_ = np.zeros(len(x))
    ones = np.ones(len(t_rg_))
    for w, r, xr_cy in zip(xr_weights, rg_params, xr_cy_list):  # weighted sum without baseline part
        if w > 0.001:
            t_rg_ += r * safe_ratios(ones, xr_cy, xr_ty, debug=False)   # using safe_ratios to avoid inconveniences by outliers

    rg_curves1 = []
    rg_curves2 = []
    segments = rg_curve.segments
    k = 0
    for state, slice_ in zip(rg_curve.states, rg_curve.slices):
        if state == 0:
            continue
        x_, y_, rg_ = segments[k]
        rg_curves1.append((x_, rg_))
        rg_curves2.append((x[slice_], t_rg_[slice_]))
        k += 1

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            fig.suptitle("compute_rg_curves debug")
            axt = ax.twinx()
            axt.grid(False)

            for cy in xr_cy_list:
                ax.plot(x, cy, ":")
            ax.plot(x, xr_ty, ":", color="red")

            for (x1_, rg1_), (x2_, rg2_) in zip(rg_curves1, rg_curves2):
                axt.plot(x1_, rg1_)
                axt.plot(x2_, rg2_, ":")
            fig.tight_layout()
            plt.show()

    return rg_curves1, rg_curves2

def plot_rg_curves(ax, xrh_params, rg_params, x, xr_cy_list, xr_ty, rg_curve):
    if len(xr_cy_list) > len(rg_params):
        xr_ty_ = xr_ty - xr_cy_list[-1]      # remove baseline component
    else:
        # assuming baseline component has been removed
        xr_ty_ = xr_ty
    weights = xrh_params/np.max(xrh_params)
    num_components = len(xrh_params)
    assert len(rg_params) == num_components

    rg_curves1, rg_curves2 = compute_rg_curves(x, weights, rg_params, xr_cy_list[:num_components], xr_ty_, rg_curve)

    k = 0
    for (x1, rg1), (x2, rg2) in zip(rg_curves1, rg_curves2):
        label = 'observed rg' if k == 0 else None
        ax.plot(x1, rg1, color='gray', alpha=0.5, label=label)
        label = 'reconstructed rg' if k == 0 else None
        ax.plot(x2, rg2, ':', color='black', label=label)
        k += 1

    ymin, ymax = ax.get_ylim()
    ax.set_ylim(min(10, ymin), max(50, ymax))
    ax.legend(loc='upper left')

def rg_deviation_inspect_impl(gdev, valid_components, rg_params, lrf_rgs):
    print("rg_deviation_inspect_impl")
    print("valid_components=", valid_components)
    print("rg_params=", rg_params)
    print("lrf_rgs=", lrf_rgs)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        ax1.set_title("Rg Plot", fontsize=16)
        ax2.set_title("Rg Curve", fontsize=16)
        ax1.plot(rg_params, "o", label='Rg params')
        ax1.plot(lrf_rgs, "o", label='LRF Rgs')
        ax1.legend()
        fig.tight_layout()
        plt.show()


"""
Computing the 3D Radial Distribution Function from Particle Positions: An Advanced Analytic Approach
Bernd A. F. Kopera and Markus Retsch
Analytical Chemistry 2018 90 (23), 13909-13914
DOI: 10.1021/acs.analchem.8b03157
"""

from scipy import spatial
from numpy import sqrt, pi, zeros, array_equal

"""
This function calculates the pair correlation function g(2)(r)
for a dense set of particles in a rectangular box. The particle
coordinates are supplied as [[x, y, z],...] lists.
All particles are used as central particles.
The spherical shell might extend beyond the known region.
A KDTree data structure is used for efficiency.
"""

"""************************************************************"""

def RDF_Simple(Particles, r, dr):

    # preallocate list for Gr
    Gr = zeros(len(r))

    # maximal radial distance
    MaxDist = r[-1] + dr/2

    # sort all Particles in a KDTree
    ParticleTree = spatial.KDTree(Particles)

    # use every particle as the center once
    k = 0
    for CentralP in Particles:

        # these are the indices for the particles at most MaxDist away:
        NNIndices = ParticleTree.query_ball_point(CentralP,
                                                  MaxDist, p = 2 , eps = 0)

        # look at every other particle at most MaxDist away:
        for Neighbour in NNIndices:

            if not array_equal(CentralP, Particles[Neighbour]):

                # calc the distance to the neighbour
                dx = CentralP[0] - Particles[Neighbour][0]
                dy = CentralP[1] - Particles[Neighbour][1]
                dz = CentralP[2] - Particles[Neighbour][2]

                d = sqrt(dx**2 + dy**2 + dz**2)

                # what bins is the particle in?
                IdxList = [k for k in range(0, len(r), 1 )
                            if abs(r[k] - d) <= dr]

                # add one to every bin the particle is in
                for Pos in IdxList:

                    # count the particle
                    Gr[Pos] += 1
        k += 1
        print("Finished Particle " + str(k) + " of " + str(len(Particles)))

    # final normalization
    Gr[0] = 0

    BoxVol = (max(Particles[0]) - min(Particles[0]))*\
             (max(Particles[1]) - min(Particles[1]))*\
             (max(Particles[2]) - min(Particles[2]))

    for i in range(1, len(r), 1):

        Gr[i] /= len(Particles) # average over all central particles

        # the most inner shells are spheres!
        if r[i] - dr/2 > 0:

            # by the shell volume
            Gr[i] /= 4/3 * pi * ((r[i] + dr/2)**3 - (r[i] - dr/2)**3)

        else:
            Gr[i] /= 4/3 * pi * (r[i] + dr/2)**3    # by the shell volume

        Gr[i] *= BoxVol / len(Particles)    # by the number density

    return Gr

"""************************************************************"""



"""
    V2Menu.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.MenuButton import MenuButton
from molass_legacy._MOLASS.Version import is_developing_version
from molass_legacy._MOLASS.SerialSettings import clear_v2_temporary_settings
from molass_legacy.Global.V2Init import update_sec_settings

class V2Menu(Tk.Frame):
    def __init__(self, parent, dialog, adjusted_sd=False):
        Tk.Frame.__init__(self, parent)
        self.parent = parent
        self.dialog = dialog
        self.adjusted_sd = adjusted_sd

        MENU_LIST = [
                    ("Optimization", lambda: self.v2_exec_wrapper(self.show_peak_editor, "V2 Optimization")),
                    ("Result Viewer", lambda: self.v2_exec_wrapper(self.show_result_viewer, "V2 Result Viewer")),
                    ("Result Animation", lambda: self.v2_exec_wrapper(self.show_result_animation, "V2 Result Animation")),
                    # ("Automated Execution", lambda: self.v2_exec_wrapper(self.show_automated_execution, "V2 Automated Execution")),
                    ]

        if is_developing_version():
            MENU_LIST.append(("Parameter Transition", lambda: self.v2_exec_wrapper(self.show_parameter_transition, "V2 parameter Transition")))

        self.menu = MenuButton(self, "▶▶ V2 Menu", MENU_LIST)
        # self.menu.entryconfig(2, state=Tk.DISABLED)
        self.menu.pack()

    def config(self, **kwargs):
        self.menu.config(**kwargs)

    def v2_exec_wrapper(self, proc, proc_name):
        self.dialog.set_state_guide_message("Executing " + proc_name)
        self.dialog.exec_wrapper(proc, proc_name)
        self.dialog.set_state_guide_message(proc_name + " done.")

    def get_ready_for_peak_editor(self):
        if not self.dialog.get_it_ready_for_qmm():
            return False

        dialog = self.dialog
        logger = dialog.analyzer.app_logger     # not yet used
        in_folder = dialog.in_folder.get()

        self.data_folder    = in_folder.replace( '\\', '/' )
        tester_info = self.dialog.tester_info
        if tester_info is None:
            tester_info_log = ''
        else:
            tester_info_log = ' with test pattern ' + str( tester_info.test_pattern )
        logger.info( "start analysis for " + self.data_folder + tester_info_log )
        return True

    def show_peak_editor(self, pe_proxy=None, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Optimizer.OptStrategyDialog
            reload(molass_legacy.Optimizer.OptStrategyDialog)
        from molass_legacy.Optimizer.OptStrategyDialog import OptStrategyDialog
        from molass_legacy.Optimizer.OptimizerUtils import show_peak_editor_impl
        clear_v2_temporary_settings()
        update_sec_settings()

        self.strategy_dialog = None
        self.peak_editor = None
        self.fullopt_dialog = None

        dialog = self.dialog
        if pe_proxy is None:
            strategy_dialog = OptStrategyDialog(dialog, dialog.serial_data, pre_recog=dialog.pre_recog)
            self.strategy_dialog = strategy_dialog
            strategy_dialog.show()
            if not strategy_dialog.applied:
                return
        else:
            strategy_dialog = None

        if not self.get_ready_for_peak_editor():
            # what happend?
            # task: clarify this case
            return

        show_peak_editor_impl(strategy_dialog, dialog, pe_proxy=pe_proxy,
                                pe_ready_cb=self.pe_ready_cb, apply_cb=self.pe_apply_cb)

        self.dialog.reset_tmp_logger("fullopt_dialog")

    def get_strategy_dialog(self):
        return self.strategy_dialog

    def pe_ready_cb(self, peak_editor):
        self.peak_editor = peak_editor

    def get_peak_editor(self):
        return self.peak_editor

    def pe_apply_cb(self, fullopt_dialog):
        self.fullopt_dialog = fullopt_dialog

    def get_fullopt_dialog(self):
        return self.fullopt_dialog

    def show_result_folder_selector(self, proc_name, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.Optimizer.ResultFolderSelector
            reload(molass_legacy.Optimizer.ResultFolderSelector)
        from molass_legacy.Optimizer.ResultFolderSelector import show_result_folder_selector_impl

        clear_v2_temporary_settings()
        update_sec_settings()

        # logger = self.dialog.analyzer.app_logger
        self.trace_dialog = None
        show_result_folder_selector_impl(self.dialog, proc_name, ready_cb=self.td_ready_cb)
        # self.dialog.reset_tmp_logger("trace_dialog")  # this results in "final_error.log" ???

    def show_result_viewer(self, debug=False):
        self.show_result_folder_selector("Result Viewer", debug=debug)

    def show_result_animation(self, debug=False):
        self.show_result_folder_selector("Result Animation", debug=debug)

    def show_parameter_transition(self, debug=False):
        self.show_result_folder_selector("Parameter Transition", debug=debug)

    def td_ready_cb(self, trace_dialog):
        self.trace_dialog = trace_dialog

    def get_trace_dialog(self):
        return self.trace_dialog
    
    def show_automated_execution(self):
        clear_v2_temporary_settings()
        update_sec_settings()



# coding: utf-8
"""
    Pomegranate.EghMixtureModel.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.interpolate import UnivariateSpline
from molass_legacy.KekLib.SciPyCookbook import smooth
from pomegranate import *
from molass_legacy.Peaks.ElutionModels import egh_pdf, compute_egh_params
from Prob.GaussianMixture import hist_to_source
from .EghDistribution import EghDistribution
import molass_legacy.KekLib.DebugPlot as plt

def sort_dist_params_lists(distributions, weights):
        components = sorted(list(zip(distributions, weights)), key=lambda x: x[0].parameters[0])
        distributions, weights = zip(*components)
        return distributions, weights

class FitResult:
    def __init__(self, model):
        distributions = model.distributions
        weights = np.exp(model.weights)     # note that weights have been converted to log
        distributions, weights = sort_dist_params_lists(distributions, weights)
        self.model = model = GeneralMixtureModel(distributions, weights)
        self.params_list = [d.parameters for d in distributions]

class EghMixtureModel:
    def __init__(self, x, y, num_components=None):
        self.logger = logging.getLogger(__name__)
        sy = smooth(y)
        self.X = np.expand_dims(hist_to_source(x, sy),1)
        self.num_components = num_components

    def fit(self, init_result=None, peaks=None):
        if init_result is None and peaks is None:
            model = GeneralMixtureModel.from_samples(EghDistribution, self.num_components, self.X)
        else:
            if init_result is None:
                distributions,weights = self.get_init_params_from_peaks(peaks)
            else:
                distributions = init_result.model.distributions
                weights = init_result.model.weights
            model = GeneralMixtureModel(distributions, weights)
            model.fit(self.X)
        return FitResult(model)

    def get_components(self, fit_result, x, y):
        model = fit_result.model
        ty = np.zeros(len(x))
        cy_list = []
        for w, dist in zip(np.exp(model.weights), model.distributions):
            tR, sigma, tau = dist.parameters
            print(w, tR, sigma, tau)
            cy = w * egh_pdf(x, tR, sigma, tau)
            cy_list.append(cy)
            ty += cy

        scale = np.max(y)/np.max(ty)
        return [scale*cy for cy in cy_list], scale*ty

    def get_init_params_from_peaks(self, rb_peaks):
        distributions = []
        heights = []
        x = rb_peaks.x
        y = rb_peaks.y
        peaks = rb_peaks.get_peaks()
        for peak in peaks:
            ls, pt, rs = peak
            tR = x[pt]
            sigma = (x[pt] - x[ls] + x[rs] - x[pt])/2
            tau = 0
            dist = EghDistribution(tR, sigma, tau)
            distributions.append(dist)
            heights.append(y[pt])

        distributions, weights = sort_dist_params_lists(distributions, heights)
        weights = np.array(weights)
        weights /= np.sum(weights)

        num_peaks = len(peaks)
        num_components = self.num_components
        num_sub_peaks = num_components - num_peaks
        assert num_sub_peaks >= 0
        if num_sub_peaks == 0:
            return distributions, weights

        spline = UnivariateSpline(x, y, s=0, ext=3)
        model = GeneralMixtureModel(distributions, weights)
        model.fit(self.X)

        taus = [d.parameters[2] for d in model.distributions]
        if abs(taus[0]) > abs(taus[-1]):
            sides = [-1, 1]
        else:
            sides = [1, -1]

        distributions = list(distributions)

        for side in sides:
            if side < 0:
                # left side
                i = 0
                dist = model.distributions[i]
                parameters = dist.parameters
                peak_tR,peak_sigma, peak_tau = parameters
                tR = peak_tR - peak_sigma*2.5
                sigma = peak_sigma/4
                tau = peak_tau/4
            else:
                # right side
                i = -1
                dist = model.distributions[i]
                parameters = dist.parameters
                peak_tR,peak_sigma, peak_tau = parameters
                tR = peak_tR + peak_sigma*2.5
                sigma = peak_sigma/4
                tau = peak_tau/4
            h = spline(tR)
            sub_dist = EghDistribution(tR, sigma, tau)
            distributions.insert(i, sub_dist)
            heights.insert(i, h)
            num_peaks += 1
            if num_peaks == num_components:
                break

        distributions, weights = sort_dist_params_lists(distributions, heights)
        weights = np.array(weights)
        weights /= np.sum(weights)

        return distributions, weights



"""
    Models.Stochastic.MonoporeStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from matplotlib.widgets import Slider, Button
from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments
from molass_legacy.Models.Stochastic.MonoporeUtils import compute_monopore_curves

def study(x, y, baseline, model, peaks, peak_rgs, props):
    from importlib import reload
    import Models.Stochastic.RoughGuess
    reload(Models.Stochastic.RoughGuess)
    from molass_legacy.Models.Stochastic.RoughGuess import guess_monopore_params_roughtly
    import Models.Stochastic.MonoporeGuess
    reload(Models.Stochastic.MonoporeGuess)
    from molass_legacy.Models.Stochastic.MonoporeGuess import guess_monopore_params_using_moments

    moments_list = compute_egh_moments(peaks)
    monopore_params = guess_monopore_params_roughtly(x, y, model, peaks, peak_rgs, props, moments_list, debug=True)
    print("peak_rgs = ", peak_rgs)
    print("monopore_params = ", monopore_params)
    if monopore_params is None:
        return

    with plt.Dp(button_spec=["OK", "Cancel"]):
        fig, (ax1,ax2) = plt.subplots(ncols=2, figsize=(16,5))
        fig.suptitle("Stochastic Monopore Model Study", fontsize=20)
        ax1.set_title("EGH Decomposition", fontsize=16)
        ax1.plot(x, y, label='data')
        cy_list = []
        for k, params in enumerate(peaks):
            cy = model(x, params)
            cy_list.append(cy)
            ax1.plot(x, cy, ":", label='component-%d' % k)
        ty = np.sum(cy_list, axis=0)
        ax1.plot(x, ty, ":", color="red", label='model total')
        ax1.plot(x, baseline, color="red", label='baseline')
        ax1.legend()

        ax2.set_title("Monopore Decomposition", fontsize=16)
        ax2.plot(x, y, label='data')
        for M in moments_list:
            m = M[0]
            s = np.sqrt(M[1])
            ax2.axvline(m, color="green")
            ax2.axvspan(m-s, m+s, color="cyan", alpha=0.2)

        cy_list, ty = compute_monopore_curves(x, monopore_params[0:6], peak_rgs, monopore_params[6:])
        artists = []
        for k, cy in enumerate(cy_list):
            curve, = ax2.plot(x, cy, ":", label='component-%d' % k)
            artists.append(curve)
        curve, = ax2.plot(x, ty, ":", color="red", label='model total')
        artists.append(curve)

        seleccted_indeces = [0,1,2,5]
        study_params = monopore_params[seleccted_indeces]
        update_params = monopore_params.copy()
        slider_specs = [    ("N", 300, 4000, study_params[0]),
                            ("T", 0, 3, study_params[1]),
                            ("t0",  -500, 500, study_params[2]),
                            ("poresize", 10, 600, study_params[3]),
                            ]
        def slider_update(k, val):
            # print([k], "slider_update", val)
            study_params[k] = val
            update_params[seleccted_indeces] = study_params
            cy_list, ty = compute_monopore_curves(x, update_params[0:6], peak_rgs, update_params[6:])
            for curve, y_ in zip(artists, cy_list + [ty]):
                curve.set_ydata(y_)
            fig.canvas.draw_idle()

        slider_axes = []
        sliders = []
        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([0.8, 0.7 - 0.08*k, 0.11, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        fig.tight_layout()
        fig.subplots_adjust(right=0.7)
        ret = plt.show()

    if not ret:
        return
    
    egh_moments_list = compute_egh_moments(peaks)
    better_monopore_params = guess_monopore_params_using_moments(x, y, egh_moments_list,
                                                                 peak_rgs, props, monopore_params, debug=True)
    return ret


"""
    Optimizer.FactorCurvePlot.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.special import iv
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Slider
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from SecTheory.BasicModels import robust_single_pore_pdf as monopore_pdf

class FactorCurvePlot(Dialog):
    def __init__(self, parent, inspect):
        self.inspect = inspect
        Dialog.__init__(self, parent, "Factor Curve Plot", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X)

        fig, axes = plt.subplots(ncols=3, figsize=(18,5))
        self.fig = fig
        self.axes = axes

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        self.draw_curves(fig, axes)

    def draw_curves(self, fig, axes):
        inspect = self.inspect
        t0, d, N, me, T, mp = inspect.get_values()
        xr_params = inspect.get_xr_params()
        rg_params = inspect.get_rg_params()
        rho = rg_params/d
        rho[rho > 1] = 1

        ax1, ax2, ax3 = axes

        x = inspect.x
        y = inspect.y
        axts = []
        for ax in axes:
            ax.plot(x, y, color="orange")
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            axt = ax.twinx()
            axt.grid(False)
            axts.append(axt)

        ones = np.ones(len(x))
        t = x - t0
        nc = len(rg_params)

        for k, r in enumerate(rho):
            np_ = N * (1 - r)**me
            tp_ = T * (1 - r)**mp
            w = xr_params[k]

            w = (k+1)/(nc+1)
            gy = ones*(ymin*w + ymax*(1-w))

            # iv(1, np.sqrt(4*np_*t/tp_))
            v = iv(1, np.sqrt(4*np_*t/tp_))
            finite = np.isfinite(v)
            ax1.plot(x[finite], gy[finite], "o", markersize=1)
            axts[0].plot(x, w * v, ":")

            # np.sqrt(np_/(t*tp_))
            v = np.sqrt(np_/(t*tp_))
            finite = np.isfinite(v)
            ax2.plot(x[finite], gy[finite], "o", markersize=1)
            axts[1].plot(x, w * v, ":")

            # np.exp(-t/tp_-np_)
            v = np.exp(-t/tp_-np_)
            finite = np.isfinite(v)
            ax3.plot(x[finite], gy[finite], "o", markersize=1)
            axts[2].plot(x, w * v, ":")

        fig.tight_layout()
        self.mpl_canvas.draw()



# coding: utf-8
"""
    DelayOptimizer.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from SerialDataUtils import get_mtd_filename
from FlowSimulator import FlowSimulator
import molass_legacy.KekLib.DebugPlot as plt

class DelayOptimizer:
    def __init__(self, in_folder, xdata):
        self.y = xdata.e_y
        self.size = xdata.data.shape[1]
        filepath = get_mtd_filename(in_folder)
        self.simulator = FlowSimulator(filepath)
        self.init_params = self.simulator.make_simulation_data(debug=False)
        self.slice_ = self.simulator.guess_range(self.size)
        self.recognize_y_channel_location()
        t, x = self.init_params[0:2]
        print('x[0:5]=', x[0:5])
        self.sm_info = self.simulator.get_linear_range_info(t)

    def get_slice(self):
        return self.slice_

    def get_initial_data(self):
        return self.init_params

    def recognize_y_channel_location(self):
        t, x = self.init_params[0:2]
        i, j = self.simulator.get_y_channel_interval(t)
        slice_ = slice(int(x[i]), int(x[j]))
        print('i=', i, j, slice_)
        y_ = self.y[slice_]
        k = np.argmax(y_)
        self.y_channel = slice_.start + k
        print('self.y_channel=', self.y_channel)

    def get_y_channel_point(self):
        return self.y_channel, self.y[self.y_channel]

    def get_optimized_data(self):
        self.optimize_with_two_steps()
        return self.opt_params

    def optimize_with_two_steps(self):
        simulator = self.simulator
        i = self.sm_info.yc_peak
        j = self.sm_info.lin_start
        k = self.sm_info.lin_end
        xi = self.init_params[1]
        k_ = int(xi[k])
        j_ = int(xi[j])
        xsize = k_ - j_
        ysize = self.y[k_] - self.y[j_]

        def delay_error(params):
            dvf = params[0]
            smt, smx, smy = simulator.make_simulation_data(delay_factor=dvf)
            i = simulator.get_y_channel_index(smt)
            c_error = ((smx[i] - self.y_channel)/xsize)**2
            return c_error

        init_params = np.array([simulator.delay_factor])
        result = minimize(delay_error, init_params, method=None, bounds=[(-10, 20)])
        opt_dvf = result.x[0]

        smt, smx, smy = simulator.make_simulation_data(delay_factor=opt_dvf)
        i, j, k = simulator.get_optimizer_indeces(smt)
        """
        to be precise, correct rounding error
        """
        xmin = smx[j]
        xmax = smx[k]
        ymin = smy[j]
        ymax = smy[k]
        j_, k_ = int(xmin), int(xmax)+1
        slope = (ymax - ymin)/(xmax - xmin)
        y0 = np.linspace(ymin+slope*(j_ - xmin), ymax+slope*(k_ - 1 - xmax), k_ - j_)
        debug = [0]
        def overlay_error(params):
            vscale = params[0]
            y1 = y0*vscale
            y2 = self.y[j_:k_]
            v_error = np.average(((y1-y2)/ysize)**2)
            if debug[0]:
                print('vscale=', vscale)
                print('v_error=', v_error)
                fig = plt.figure()
                ax = fig.gca()
                x = np.arange(j_, k_)
                ax.plot(x, y1)
                ax.plot(x, y2, color='orange')
                ax.plot(smx[i-1:], smy[i-1:]*vscale, color='blue')  # this line must overlap with y1 line
                ax.plot(smx[i], smy[i]*vscale, 'o', color='red')
                ax.plot(self.y_channel, self.y[self.y_channel], 'o', color='yellow')
                ret = plt.show()
                if not ret:
                    debug[0] = 0

            return v_error

        init_endy = self.init_params[2][k]
        init_vscale = ysize/init_endy
        print('ysize=', ysize, 'init_endy=', init_endy)
        init_params = np.array([init_vscale])
        print('init_params=', init_params)
        method = 'L-BFGS-B'
        # method = 'TNC'
        # method = 'SLSQP'
        # method = 'Nelder-Mead'
        result = minimize(overlay_error, init_params, method=method, bounds=[(init_vscale*0.5, init_vscale*1.5)])
        vscale = result.x[0]
        print('opt_dvf, vscale=', opt_dvf, vscale)

        sm_info = simulator.get_linear_range_info(smt, vscale)
        self.opt_params = smt, smx, smy, sm_info



"""
    ATSAS.AutorgRunner.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.AutorgKek.AtsasTools import autorg
from molass_legacy._MOLASS.WorkUtils import get_temp_folder

class AutorgRunner:
    def __init__(self):
        self.temp_folder = get_temp_folder()

    def run_from_array(self, data, remove_tempfile=True):
        tempfile = os.path.join(self.temp_folder, 'autorg-%d.txt' % os.getpid())
        np.savetxt(tempfile, data)
        orig_result, eval_result = autorg(tempfile)
        if remove_tempfile:
            os.remove(tempfile)
        return orig_result, eval_result


"""
    Solvers.ABC.CustomDistance.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from typing import Callable, Union
from pyabc.distance.base import Distance, NoDistance

class CustomFunctionDistance(Distance):
    """
    This is a wrapper around a simple function which calculates the distance.
    If a function/callable is passed to the ABCSMC class, which is not
    subclassed from pyabc.Distance, then it is converted to an instance of the
    SimpleFunctionDistance class.

    Parameters
    ----------
    fun: Callable[[dict, dict], float]
        A Callable accepting as parameters (a subset of) the arguments of the
        pyabc.Distance.__call__ function. Usually at least the summary
        statistics x and x_0. Returns the distance between both.
    """

    def __init__(self, fun):
        super().__init__()
        self.fun = fun

    def __call__(
        self,
        x: dict,
        x_0: dict,
        t: int = None,
        par: dict = None,
    ) -> float:
        return self.fun(x, x_0, t, par)

    def get_config(self):
        conf = super().get_config()
        # try to get the function name
        try:
            conf["name"] = self.fun.__name__
        except AttributeError:
            try:
                conf["name"] = self.fun.__class__.__name__
            except AttributeError:
                pass
        return conf

    @staticmethod
    def to_distance(maybe_distance: Union[Callable, Distance]) -> Distance:
        """
        Parameters
        ----------
        maybe_distance: either a Callable as in FunctionDistance, or a
        pyabc.Distance object.

        Returns
        -------
        A Distance instance.
        """
        if maybe_distance is None:
            return NoDistance()

        if isinstance(maybe_distance, Distance):
            return maybe_distance

        return CustomFunctionDistance(maybe_distance)




# coding: utf-8
"""

    SerialControlInfo.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF

"""
import os

class SerialControlInfo:
    def __init__( self,
                    data_folder,
                    conc_folder,
                    work_folder,
                    temp_folder,
                    conc_file=None,
                    serial_data=None,
                    guinier_folder=None,
                    stamp_file=None,
                    serial_file='serial_result.csv',
                    book_file='serial_analysis_report.xlsx',
                    min_analysis_value=0.5,
                    range_type=0,
                    analysis_ranges=None,
                    preview_params=None,
                    zx=False,
                    mapped_info=None,
                    maintenance_log=None,
                    env_info=None,
                    known_info_list=None,
                    cleaner=None,
                    parent=None ):

        self.data_folder    = data_folder
        self.conc_folder    = conc_folder
        self.work_folder    = work_folder
        self.temp_folder    = temp_folder
        self.conc_file      = conc_file
        self.serial_data    = serial_data
        self.guinier_folder = guinier_folder
        self.stamp_file     = stamp_file
        self.serial_file    = serial_file
        self.book_file      = book_file
        self.min_analysis_value    = min_analysis_value
        self.range_type     = range_type
        self.analysis_ranges = analysis_ranges
        self.preview_params = preview_params
        self.zx             = zx
        self.mapped_info    = mapped_info
        self.maintenance_log    = maintenance_log
        self.env_info       = env_info
        self.known_info_list = known_info_list
        self.cleaner        = cleaner
        self.parent         = parent
        self.more_multicore = os.cpu_count() > 4
        # self.more_multicore = False



"""
    ResultGui.py

    Copyright (c) 2016-2022, SAXS Team, KEK-PF
"""

import os
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.KekLib.TkUtils import adjusted_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.AutorgKek.AtsasTools import autorg as autorg_atsas
# from KekToolsGP             import autorg   as autorg_kekpf
from molass_legacy.GuinierAnalyzer.AutorgKekAdapter import AutorgKekAdapter
from molass_legacy.SerialAnalyzer.ProofPlot import Plotter
from molass_legacy.SerialAnalyzer.ResultTable import ResultTable
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.TextShowDialog import TextShowDialog
from molass_legacy.KekLib.NumpyUtils import np_savetxt

class ResultGui( Tk.Toplevel ):
    def __init__( self, parent, folder, file, np_array=None, atsas_exe_index=0,
                    orig_info=None,
                    q_limit_index=None,
                    title_filename=None,
                    atsas_np_array=None ):
        Tk.Toplevel.__init__( self, parent )

        self.orig_info      = orig_info
        self.q_limit_index  = q_limit_index
        self.atsas_np_array = atsas_np_array

        self.withdraw()
        tk_set_icon_portable( self )

        self.oopt_fit_consistency   = get_setting( 'oopt_fit_consistency' )
        # self.oopt_qrg_limits        = get_setting( 'oopt_qrg_limits' )
        self.oopt_qrg_limits        = 'Limited'
        # self.oopt_qrg_limits_apply  = get_setting( 'oopt_qrg_limits_apply' )
        self.oopt_qrg_limits_apply  = 1
        # self.oopt_optimize_only     = get_setting( 'oopt_optimize_only' )
        self.oopt_optimize_only     = 0

        rec_array, intensity = self.run_autorg(  folder, file, np_array=np_array, atsas_exe_index=atsas_exe_index )
        self.intensity = intensity

        atsas_exe_paths = get_setting( 'atsas_exe_paths' )
        if len( atsas_exe_paths ) > 0:
            text_ = 'KEK autorg result of %s, compared with that of %s' % (
                        folder + '/' + file,
                        atsas_exe_paths[atsas_exe_index]
                        )

        else:
            text_ = 'KEK autorg result of %s' % ( folder + '/' + file )

        frame = Tk.Frame( self )
        frame.pack( fill=Tk.BOTH, expand=1, padx=10, pady=10 )

        frame1 = Tk.Frame( frame )
        frame1.pack( fill=Tk.BOTH, expand=1 )

        """
        if self.oopt_fit_consistency == 1:
            text_ += '; optimized'
            if self.oopt_qrg_limits=='Limited' and self.oopt_qrg_limits_apply:
                text_ += ' with q*Rg limits'
            if self.oopt_optimize_only == 1:
                text_ += '; optimized only'
        """

        label = Tk.Label( frame1, text=text_ )
        label.pack()

        frame2 = Tk.Frame( frame )
        frame2.pack( fill=Tk.BOTH, expand=1 )

        self.show_results( frame1, rec_array )

        if title_filename is None:
            title_filename = file
        self.show_proof_plot( frame2, title_filename, rec_array, intensity )

        # it seemd that '<Control-c>' should be bound from the top-level.
        self.bind( "<Control-c>", self.control_c )

        # print( 'ResultGui: geometry=', self.geometry() )
        self.update()
        self.geometry( adjusted_geometry( '1800x700' ) )

        if intensity.comments is not None:
            self.create_popup_menu()
            self.bind( '<Button-3>', self._on_right_click )

        self.protocol( "WM_DELETE_WINDOW", self.at_exit )
        self.deiconify()

    def create_popup_menu( self ):
        # create a popup menu
        self.menu = Tk.Menu( self, tearoff=0 )
        self.menu.add_command( label='Show Comments',  command=self.show_comments )

    def _on_right_click( self, event ):
        self.popup_position = ( event.x, event.y )
        # print( 'popup_position=', self.popup_position )

        self.menu.post( event.x_root, event.y_root )

    def show_comments( self ):
        dialog = TextShowDialog( self, "Intensity File Header",
                message=self.intensity.comments,
                width=93, height=53,
                )

    def run_autorg( self, folder, file, np_array=None, atsas_exe_index=0 ):
        path = folder + '/' + file

        optimize_           = self.oopt_fit_consistency==1

        if self.oopt_qrg_limits == 'Unlimited':
            qrg_limits_ = None
            qrg_limits_apply_   = False
        else:
            qrg_limits_ = get_setting( 'oopt_qrg_limits_vals' )
            qrg_limits_apply_   = self.oopt_qrg_limits_apply == 1

        # print( 'ResultGui.run_autorg: qrg_limits_=', qrg_limits_ )

        rec_array = []
        atsas_exe_paths = get_setting( 'atsas_exe_paths' )
        if len( atsas_exe_paths ) > 0 and atsas_exe_index is not None:
            if self.atsas_np_array is None:
                path_for_atsas  = path
            else:
                path_for_atsas  = get_setting( 'temp_folder' ) + '/' + file
                np_savetxt( path_for_atsas, self.atsas_np_array )
            rg_info_a, rg_info_a_   = autorg_atsas( path_for_atsas, exe_index=atsas_exe_index, smaxrg=qrg_limits_[1] )
            if self.atsas_np_array is not None:
                os.remove( path_for_atsas )
        else:
            rg_info_a, rg_info_a_   = None, None

        rec_array.append( rg_info_a )
        rec_array.append( rg_info_a_ )

        if np_array is None:
            path_for_autorg_kek = path
        else:
            path_for_autorg_kek = np_array

        """
        rg_info_k, intensity = autorg_kekpf( path_for_autorg_kek, robust=True,
                                optimize=optimize_,
                                qrg_limits=qrg_limits_,
                                qrg_limits_apply=qrg_limits_apply_ )
        """

        autorg_kek  = AutorgKekAdapter( path_for_autorg_kek )
        result_kek  = autorg_kek.run( robust=True, optimize=True, fit_result=True )

        rec_array.append( result_kek )
        intensity = autorg_kek.intensity

        return rec_array, intensity

    def show_proof_plot( self, frame, file, rec_array, intensity ):
        rg_info_a   = rec_array[1]
        rg_info_k   = rec_array[2]
        plotter = Plotter( intensity, file, rg_info_a, rg_info_k, orig_info=self.orig_info, q_limit_index=self.q_limit_index )
        plotter.show_prepare()

        self.mpl_canvas = FigureCanvasTkAgg( plotter.fig, frame )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

    def show_results( self, frame, rec_array ):
        self.result = ResultTable( frame, rec_array)

    def at_exit( self ):
        # print( 'at_exit' )
        self.destroy()

    def control_c( self, event ):
        w = self.focus_get()
        # print( 'table=', self.file_info_table )
        # print( 'w=', w )
        if str(w).find( str( self.result ) ) == 0:
            self.result.table.control_c( event )
        else:
            # TODO:  w.selection
            pass



"""
    OptJobInfo.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import os
from .OptLogFile import OptLogFile

JOB_STATES = ["to be run", "running", "finished", "dead"]

class OptJobInfo:
    def __init__(self, name=None, state=None, nc=None, drift_type=None, niter=None, seed=None, pid=None):
        self.name = name
        self.state = state
        self.nc = nc
        self.drift_type = drift_type
        self.niter = niter
        self.seed = seed
        self.pid = pid
        self.items  = [name, state, nc, drift_type, niter, seed]    # for backward compatibility

    def __repr__(self):
        return 'OptimizerJob' + str(self.items)

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, item):
        return self.items[item]

    def update_state(self, state):
        self.state = state

    def load(self, folder):
        folder_name = os.path.split(folder)[1]
        if self.name is None:
            self.name = folder_name
        else:
            assert folder_name == self.name

        path = os.path.join(folder, "optimizer.log")
        log_file = OptLogFile(path)
        optdict = log_file.optdict

        self.state = "done"
        self.nc = int(optdict['-n'])
        self.drift_type = optdict['-d']
        self.niter = int(optdict['-m'])
        self.seed = int(optdict['-s'])
        self.pid = None
        self.items  = [self.name, self.state, self.nc, self.drift_type, self.niter, self.seed]
        # self.items will be reprecated



"""
    DefaultNumPeaks.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.Optimizer.OptConstants import MAX_NUM_PEAKS

def get_default_num_peaks(sd):
    try:
        xr_num_peaks = len(sd.get_xray_curve().peak_info)
        uv_num_peaks = len(sd.get_uv_curve().peak_info)
        if xr_num_peaks == uv_num_peaks:
            num_peaks = xr_num_peaks + 2
        else:
            num_peaks = max(xr_num_peaks, uv_num_peaks) + 1
        ret_num_peaks = min(MAX_NUM_PEAKS, num_peaks)
    except:
        # AttributeError: 'NoneType' object has no attribute 'get_xray_curve'
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None, "get_default_num_peaks failed: ", n=5)
        raise RuntimeError

    return ret_num_peaks



"""

    PersistentInfo.py

    Copyright (c) 2016-2023, Masatsuyo Takahashi, KEK-PF

"""
import sys
import os
import pickle
import shutil
if sys.version_info > (3,):
    read_open_mode  = 'rb'
    write_open_mode = 'wb'
else:
    read_open_mode  = 'r'
    write_open_mode = 'w'

from molass_legacy.KekLib.BasicUtils     import home_dir, exe_name

def kek_tools_folder():
    return '%s/.KekTools' % ( home_dir() )

def get_pickle_file_path( name ):
    return '%s/%s/%s' % ( kek_tools_folder(), exe_name(), name )

def make_pickle_file_path( folder, file ):
    if not os.path.exists( folder ):
        os.makedirs( folder )
    app_folder   = '%s/%s' % ( folder, exe_name() )
    if not os.path.exists( app_folder ):
        os.makedirs( app_folder )

    file_path = '%s/%s' % ( app_folder, file )
    if os.path.exists(file_path):
        # check if the file is writable
        assert os.access(file_path, os.W_OK)
    else:
        # check if the folder is writable
        try:
            access_test_file = os.path.join(app_folder, 'access_test')
            fh = open(access_test_file, 'w')
            fh.close()
            os.remove(access_test_file)
        except Exception as exc:
            print('access error in', app_folder, exc)
            raise

    return file_path

class PersistentInfo:
    def __init__( self, filename, defaults={}, pickle_file=None, alt_folder=None ):
        self.alt_folder = None

        if pickle_file is None:
            try:
                if alt_folder is None:
                    folder = kek_tools_folder()
                else:
                    folder = alt_folder

                pickle_file = make_pickle_file_path(folder, filename )
                self.alt_folder = alt_folder
            except:
                try:
                    import molass_legacy.KekLib.CustomMessageBox as MessageBox
                except:
                    import OurMessageBox as MessageBox
                try:
                    alt_folder = __file__
                    for k in range(3):
                        alt_folder, _ = os.path.split(alt_folder)
                        print([k], alt_folder)
                    alt_folder = os.path.join(alt_folder, '.KekTools')
                    pickle_file = make_pickle_file_path(alt_folder, filename )
                    MessageBox.showwarning( 'Alternative Settings Folder Notification',
                        'Using %s\n'
                        'as your settings folder '
                        'due to an unidentified error.\n'
                        % alt_folder
                        )
                    self.alt_folder = alt_folder
                except:
                    MessageBox.showerror( 'No other alternative',
                        'Failed to use %s\n' % alt_folder
                        )
                    assert False

        self.pickle_file = pickle_file

        # パラメータのロード
        if os.path.exists( self.pickle_file ) and os.path.getsize( self.pickle_file ):
            try:
                pf = open( self.pickle_file, read_open_mode )
                self.dictionary = pickle.load( pf )
                pf.close()
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                etb = log_exception(None, "open( self.pickle_file): ", n=10)

                pf.close()
                pickle_dir = os.path.dirname( self.pickle_file )

                if etb.last_lines().find("AttributeError: Can't get attribute") >= 0:
                    # temporary fix to avoid inconvenience for users with a message like the following
                    # AttributeError: Can't get attribute 'FlowChange' on <module 'Trimming.FlowChange' from '...\\FlowChange.py'>
                    # consider to remove this cause
                    pass
                else:
                    try:
                        import molass_legacy.KekLib.CustomMessageBox as MessageBox
                    except:
                        import OurMessageBox as MessageBox
                    MessageBox.showwarning( 'Broken Settings Notification',
                        'Your settings infomation previously stored in\n'
                        '"%s" seems to have been broken.\n'
                        '\n'
                        'Sorry, your entire settings information will be lost\n'
                        'and reset to the default.\n'
                        '\n'
                        'This inconvenience can happen in an invocation just after such cases as listed.\n'
                        '  - abnormal termination of the program\n'
                        '  - multiple simutaneous invocation of the program\n'
                        % ( self.pickle_file ) )
                        # TODO: set parent
                try:
                    shutil.rmtree( pickle_dir )
                except:
                    # continue anyway
                    pass
                self.dictionary = {}
        else:
            self.dictionary = {}

        for k in defaults.keys():
            if self.dictionary.get(k) == None:
                self.dictionary[k] = defaults[k]

    def set_dictionary( self, dictionary ):
        self.dictionary = dictionary

    def get_dictionary( self ):
        return self.dictionary

    def save( self, file=None ):
        if file is None:
            file = self.pickle_file
        pf = open( file, write_open_mode )
        pickle.dump( self.dictionary, pf )
        pf.close()



"""
    Simulative.SimulationCushion.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.Version import is_developing_version

def lnp_model_interactive(lrf_src, guess_info, parent):
    from importlib import reload
    import Simulative.SimulationMain
    reload(Simulative.SimulationMain)
    from Simulative.SimulationMain import lognormalpore_model_interactive    
    lognormalpore_model_interactive(lrf_src, guess_info, parent)

def approx_boundary(lrf_src, guess_info, parent):
    from importlib import reload
    import Simulative.ApproxBoundary
    reload(Simulative.ApproxBoundary)
    from Simulative.ApproxBoundary import demo_impl
    x = lrf_src.xr_x
    y = lrf_src.xr_y
    demo_impl(x, y, *guess_info, parent)

def several_studies(lrf_src, guess_info, parent):
    from importlib import reload
    import Simulative.SeveralStudies
    reload(Simulative.SeveralStudies)
    from Simulative.SeveralStudies import several_studies_impl    
    several_studies_impl(lrf_src, guess_info, parent)

def temporary_trial(lrf_src, guess_info, parent):
    from importlib import reload
    import Simulative.TemporaryTrial
    reload(Simulative.TemporaryTrial)
    from Simulative.TemporaryTrial import temporary_trial_impl    
    temporary_trial_impl(lrf_src, guess_info, parent)

def demo_cushion(lrf_src, guess_info, parent=None):
    if is_developing_version():
        extra_button_specs = [
            ("LNP Interactive", lambda: lnp_model_interactive(lrf_src, guess_info, parent)),
            ("Approx Boundary", lambda: approx_boundary(lrf_src, guess_info, parent)),
            ("Several Studies", lambda: several_studies(lrf_src, guess_info, parent)),
            ("Temporary Trial", lambda: temporary_trial(lrf_src, guess_info, parent)),
            ]
        lrf_src.draw(extra_button_specs=extra_button_specs)
    else:
        lnp_model_interactive(lrf_src, guess_info, parent)


"""
    DataRange.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from bisect import bisect_right
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.Tools.OutlineViewer import (format_coord, using_normal_conc,
                OutlineViewerDialog,
                OutlineViewerFrame, OutlineViewer3dFrame, OutlineViewer2dFrame)
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, reset_setting
from .TrimmingInfo import TrimmingInfo

class DataRangeDialog(OutlineViewerDialog):
    def __init__(self, parent, pre_recog):
        self.parent = parent
        self.pre_recog = pre_recog
        self.logger = logging.getLogger( __name__ )
        self.serial_data = pre_recog.get_pre_recog_copy()
        self.pre_rg = pre_recog.pre_rg
        self.cs = pre_recog.cs
        self.using_normal_conc = using_normal_conc()
        self.using_xray_conc = get_setting('use_xray_conc') == 1
        self.xray_only = False
        Dialog.__init__(self, parent, "Data Range Trimming", visible=False)

    def get_frame_class(self):
        return DataRestrictorFrame

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, padx=50, pady=5)
        for k in range(3):
            box.columnconfigure(k, weight=1)

        self.at_a_glance_btn = Tk.Button(box, text="At a Glance", command=self.show_it_at_a_glance)
        self.at_a_glance_btn.grid(row=0, column=0)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.grid(row=0, column=1)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.grid(row=0, column=2)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def show_it_at_a_glance(self):
        from .TrimmingResult import TrimmingResultDialog
        dialog = TrimmingResultDialog(self, self.pre_recog)
        dialog.show()

    def apply(self):
        xr_restrict_list  = self.frames[0].get_info_list()
        if self.using_xray_conc:
            uv_restrict_list    = xr_restrict_list
        else:
            uv_restrict_list    = self.frames[1].get_info_list()
        self.logger.info('restriction info for xray has been set as ' + str(xr_restrict_list) )
        self.logger.info('restriction info for uv has been set as ' + str(uv_restrict_list) )
        set_setting('xr_restrict_list' , xr_restrict_list )
        set_setting('uv_restrict_list' , uv_restrict_list )
        set_setting('manually_trimmed' , True )

        reset_setting('zx_penalty_matrix')
        reset_setting('ignore_all_bqs')
        reset_setting('ignore_bq_list')
        reset_setting('range_editor_info')
        reset_setting('editor_ranges')
        reset_setting('editor_model')
        reset_setting('decomp_editor_info')

    def exec_callback(self, callback, path):
        callback(self, path)

class DataRestrictorFrame(OutlineViewerFrame):
    def __init__(self, parent, dialog, titles, x, y, data, epos, ecurve, el_color, restrict_list,
                    toggle_info, disable_x_selection=False,
                    averaged_ridge=False,
                    changeable_yscale=False, sub_curve_info=None, pre_recog=None,
                    x_axis_ends=None,
                    bridge_slice=None,
                    sub_curve_info_list=None):
        # print('x.shape=', x.shape, 'y.shape=', y.shape, 'data.shape=', data.shape)

        self.dialog = dialog
        self.data = data
        self.x  = x
        self.y  = y

        self.epos = epos
        self.ecurve = ecurve
        self.el_color = el_color
        self.restrict_list = restrict_list
        self.toggle_info = toggle_info
        self.using_xray_conc = get_setting('use_xray_conc') == 1
        self.using_normal_conc = using_normal_conc()

        Tk.Frame.__init__(self, parent)
        # self.build_body(titles, disable_x_selection, averaged_ridge, changeable_yscale, sub_curve_info, pre_recog, bridge_slice, sub_curve_info_list)
        # why self.disable_selection?
        self.build_body(titles, False, averaged_ridge, changeable_yscale, sub_curve_info, pre_recog, bridge_slice, sub_curve_info_list)

    def get_frame_classes(self):
        return OutlineViewer3dFrame, DataRestrictor2dFrame

class DataRestrictor2dFrame(OutlineViewer2dFrame):
    """
        Note that this class is used in the following four ways.
        UV elution          X elution
        UV spectral         X angular
    """
    def __init__(self, parent, parent_obj, figsize, title, x, y, data, color, opos, ocolor, restrict_info,
                sub_curve_info=None, pre_recog=None, ecurve=None, opos_sub=None, disable_selection=False,
                changeable_yscale=False,
                bridge_slice=None,
                toolbar=False, val_label=False):
        self.parent_obj = parent_obj
        self.logger = parent_obj.dialog.logger
        self.sd = parent_obj.dialog.serial_data
        self.title = title
        self.drawing_uv = title.find("UV") >= 0         # unify with showing_uv?
        self.drawing_elution = title.find("Elution") >= 0
        self.x = x
        self.y = y
        self.data = data
        self.color = color
        self.opos = opos
        self.ocolor = ocolor
        self.restrict_info = restrict_info
        self.sub_curve_info = sub_curve_info
        self.showing_X = title.find("X") >= 0 
        self.pre_recog = pre_recog
        self.ecurve = ecurve
        self.showing_uv = pre_recog is not None
        self.opos_sub = opos_sub
        self.disable_selection = disable_selection
        self.selection_guide_text = "Drag to suggest where to trim."
        self.changeable_yscale = changeable_yscale
        self.bridge_slice = bridge_slice
        self.toolbar_flag = toolbar
        self.val_label_flag = val_label
        self.max_value = len(y) - 1
        self.popup_menu = None
        self.inspector = None
        Tk.Frame.__init__(self, parent)

        self.normal_plot = title.find('Xray') >= 0 or using_normal_conc() or title.find('Elution') >= 0
        self.cbvar = None

        cframe = Tk.Frame(self)
        cframe.pack()
        self.build_cframe(cframe, figsize)
        self.add_popup_menu_bind()

        bframe = Tk.Frame(self)
        bframe.pack(fill=Tk.X, expand=1)

        if self.toolbar_flag:
            bframe_ = self.add_toolbar(bframe)
        else:
            bframe_ = bframe

        self.build_bframe(bframe_)

        if self.drawing_uv and not self.drawing_elution:
           self.draw_flat_wavelength()

        self.gra_dialog = None

    def onselect(self, xmin, xmax):
        width = xmax - xmin
        if width < self.span_ack_min_width:
            # should better be handled in SpanSelector
            # print("this is not a span selection. ignore width=", width)
            return

        i = min(len(self.x)-1, bisect_right(self.x, xmin))
        j = min(len(self.x)-1, bisect_right(self.x, xmax))

        prev_lower = self.lower.get()
        prev_upper = self.upper.get()

        self.lower.set( i )
        self.upper.set( j )
        self.cbvar.set( 1 )     # this line must be executed after the above lines execution

        self.logger.info("range changed from (%d, %d) to (%d, %d)", prev_lower, prev_upper, i, j)

    def draw_selection(self, ax, restrict_info):
        self.draw_selection_guide(ax)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        if restrict_info is None:
            if self.cbvar is not None and self.cbvar.get() == 1:
                try:
                    restrict_info = [ var.get() for var in self.ivars  ]
                except:
                    # ?
                    restrict_info = None
        if restrict_info is not None:
            for i in restrict_info:
                x_ = self.x[i]
                ax.plot( [x_, x_], [ymin, ymax], color='yellow' )

    def draw_selection_guide(self, ax):
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        tx = (xmin + xmax)/2
        if self.changeable_yscale and self.log_scale:
            # ty = np.power(10, (-2 + np.log10(ymax))/2 )
            ty = np.power(10, np.log10(ymin)*0.1 + np.log10(ymax)*0.9)
        else:
            ty = (ymin + ymax)/2
        ax.text(tx, ty, self.selection_guide_text, ha="center", va="center", alpha=0.2, fontsize=30)

    def build_bframe(self, bframe):

        btn_frame = Tk.Frame(bframe)
        # btn_frame.pack(side=Tk.RIGHT, padx=50 if self.val_label_flag else 110)
        btn_frame.pack(side=Tk.RIGHT, padx=50)

        cb_state = Tk.DISABLED if self.disable_selection or not self.normal_plot else Tk.NORMAL

        if self.restrict_info is None:
            cbvar_init = 0
            start   = None
            stop    = None
        else:
            cbvar_init = self.restrict_info.flag
            start   = self.restrict_info.start
            stop    = self.restrict_info.stop
        lower_init = 0 if start is None else start
        upper_init = self.max_value if stop is None else stop - 1

        self.logger.info("debug info: len(x)=%d, lower_init=%d, upper_init=%d", len(self.x), lower_init, upper_init)
        if upper_init > self.max_value:
            # as in 20170209/OAGIwyatt_01
            # the reason of which remains to be investigated
            self.logger.warning("forced upper_init(%d) to be equal to max_value(%d)", upper_init, self.max_value)
            upper_init = self.max_value

        self.cbvar  = Tk.IntVar()
        self.cbvar.set( cbvar_init )
        cb  = Tk.Checkbutton( btn_frame, text="trim",
                                variable=self.cbvar, state=cb_state )
        cb.grid( row=0, column=0 )

        self.cbvar.trace( 'w', lambda *args: self.cb_tracer() )

        self.cb_depentents = []
        self.lower  = Tk.IntVar()
        self.lower.set( lower_init )
        self.upper  = Tk.IntVar()
        self.upper.set( upper_init )
        self.ivars  = [ self.lower, self.upper ]

        if self.val_label_flag:
            self.val_vars = []

        colbase = 1
        num_parts = 3 if self.val_label_flag else 2
        space_count = 0
        for i, rec in enumerate( [ [ 'start', lower_init], [ 'end' , upper_init ] ] ):
            text, value = rec

            item_base = i*num_parts
            space = Tk.Frame( btn_frame, width=10 )
            space.grid( row=0, column=colbase + space_count + item_base, sticky=Tk.W )
            space_count += 1

            btn_label   = Tk.Label( btn_frame, text=text )
            btn_label.grid( row=0, column=colbase + item_base + space_count, sticky=Tk.W )

            self.cb_depentents.append( btn_label )
            entry   = Tk.Spinbox( btn_frame, textvariable=self.ivars[i],
                        from_=0, to=self.max_value, increment=1, 
                        justify=Tk.CENTER, width=6 )
            entry.grid( row=0, column=colbase + item_base + 1 + space_count, sticky=Tk.W )
            self.cb_depentents.append( entry )

            if self.val_label_flag:
                val_var = Tk.StringVar()
                val_var.set( '%.3g' % self.x[value])
                self.val_vars.append( val_var )
                val_label   = Tk.Label( btn_frame, textvariable=val_var )
                val_label.grid( row=0, column=colbase + item_base + 2 + space_count, sticky=Tk.W )
                self.cb_depentents.append( val_label )

            self.ivars[i].trace( 'w', self.spinbox_tracer )

        self.cb_tracer()    # for intial state

        if self.ecurve is not None:
            last_column = colbase + num_parts*2 + space_count
            data_name = "UV" if self.showing_X else "Xray"
            self.from_opposite_side_btn = Tk.Button(btn_frame, text="Set from " + data_name, command=self.set_from_opposite_data_info)
            self.from_opposite_side_btn.grid(row=0, column=last_column + 1, padx=10)

    def cb_tracer(self):
        cb_val = self.cbvar.get()
        state   = Tk.NORMAL if cb_val== 1 else Tk.DISABLED
        for j, w in enumerate(self.cb_depentents ):
            w.config( state=state )
            if cb_val == 1 and j == 1:
                w.focus_force()     # set focus to the spinbox
        if self.normal_plot:
            self.draw_restrict_lines( state == Tk.NORMAL )

    def set_from_opposite_data_info(self):
        dialog = self.parent_obj.dialog

        cs = dialog.cs
        A, B = cs.slope, cs.intercept

        if self.showing_X:
            A_, B_ = 1/A, -B/A
        else:
            A_, B_ = A, B

        def convert(x):
            return A_*x + B_

        oframe = dialog.get_opposite_frame()
        ocanvas = oframe.canvases[1]
        r_info = ocanvas.get_restrict_info(fill=True)

        xmin = convert(r_info.start)
        xmax = convert(r_info.stop - 1)
        self.onselect(xmin, xmax)

    def draw_restrict_lines(self, restrict):
        if restrict:
            try:
                restrict_info   = [ var.get() for var in self.ivars ]
                if self.val_label_flag:
                    for k, i in enumerate(restrict_info):
                        self.val_vars[k].set( '%.3g' % self.x[i] )
            except:
                # when entry value is deleted
                return
        else:
            restrict_info   = None

        self.draw(restrict_info=restrict_info)

    def spinbox_tracer( self, *args ):
        try:
            self.draw_restrict_lines( True )
        except:
            # when the Spinbox entry text is deleted
            pass

    def get_restrict_info(self, fill=False):
        cbval   = self.cbvar.get()
        if cbval:
            lower   = self.lower.get()
            upper   = self.upper.get()
            if lower == 0 and upper == self.max_value and not fill:
                info = None
            else:
                info = TrimmingInfo(1, lower, upper+1, len(self.x))
        else:
            if fill:
                info = TrimmingInfo(1, 0, self.max_value+1, len(self.x))
            else:
                info = None
        return info

    def add_popup_menu_bind(self):
        xmin, xmax = self.ax.get_xlim()
        self.span_ack_min_width = (xmax - xmin) * 0.05
        self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def on_figure_click(self, event):
        if event.button == 1:
            return
        elif event.button == 3:
            if get_setting('enable_gr_analyzer') and event.inaxes == self.ax:
                if self.showing_X:
                    if self.ecurve is None:
                        menu_specs = [  ("Guinier Region Inspector", self.show_guinier_inspector),
                                        ("Gradual Region Analysis", self.show_gradual_analysis),
                                        ]
                    else:
                        menu_specs = [("Show Extra Info", self.show_extra_info)]
                    self.show_popup_menu(event, menu_specs)
                return

    def show_popup_menu(self, event, menu_specs):
        from molass_legacy.KekLib.TkUtils import split_geometry
        self.create_popup_menu(event, menu_specs)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu(self, event, menu_specs):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0 )
            for spec in menu_specs:
                self.popup_menu.add_command(label=spec[0], command=spec[1])

    def show_extra_info(self):
        from .TrimmingInspection import ElutionTrimmingDialog
        parent_frame = self.parent_obj
        dialog = parent_frame.dialog
        data = parent_frame.data
        self.inspector = ElutionTrimmingDialog(dialog, dialog, data, self.ecurve, self.restrict_info)
        self.inspector.show()

    def show_guinier_inspector(self):
        from .GuinierRegion import GuinierRegionInspector
        parent_frame = self.parent_obj
        dialog = parent_frame.dialog
        data = parent_frame.data
        i_smp = self.opos       # i_smp : index of SMP - Standard Mapping Plane
        trimming_info = dialog.frames[0].get_info_list()[0]
        self.inspector = GuinierRegionInspector(dialog, dialog, data, i_smp, trimming_info.get_slice())
        self.inspector.show()

    def create_gradual_analysis(self, debug=True):
        if debug:
            from importlib import reload
            import Trimming.GradualRegionAnalysis
            reload(Trimming.GradualRegionAnalysis)
        from .GradualRegionAnalysis import GradualRegionAnalysis
        parent_frame = self.parent_obj
        dialog = parent_frame.dialog
        dialog.config(cursor='wait')
        dialog.update()
        # note that self.pre_recog can be None unlike dialog.pre_recog
        # it must be dialog.pre_recog here is stead of self.pre_recog
        trimming_info = dialog.frames[0].get_info_list()[1]     # angular slice
        gra_dialog = GradualRegionAnalysis(dialog, dialog, self.sd, trimming_info, dialog.pre_recog)
        dialog.config(cursor='')
        dialog.update()
        return gra_dialog

    def show_gradual_analysis(self):
        # separate constructor to make it possible to manipulate from the tester
        self.gra_dialog = self.create_gradual_analysis()

        self.gra_dialog.show()

    def has_gra_dialog(self):
        return self.gra_dialog is not None

    def draw_flat_wavelength(self):
        from molass_legacy.UV.PlainCurveUtils import get_flat_wavelength
        # assuming the ylim has been already set
        ax = self.ax

        ymin, ymax = ax.get_ylim()
        w = get_flat_wavelength(self.sd.lvector)
        ax.plot([w, w], [ymin, ymax], color="cyan")



"""
    V2PropOptimizer.V1PreviewAdapter.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.Mapping.MapperConstructor import create_mapper
from molass_legacy.SerialAnalyzer.AnalyzerUtil import compute_conc_factor_util
from Extrapolation.PreviewData import PreviewOptions

class ZxDialogProxy:
    def __init__(self, parent):
        self.parent = parent

class JudgeHolderProxy:
    def __init__(self, optimizer):
        peak_top_x = optimizer.xr_curve.peak_top_x.copy()
        cd_colors = ['green'] * len(peak_top_x)
        set_setting('mapper_cd_color_info', (peak_top_x, cd_colors))

class MapperProxy:
    def __init__(self, optimizer):
        self.x_curve = optimizer.xr_curve       # better be corrected
        self.optimizer = optimizer

    def get_conc_vector(self, conc_factor):
        optimizer = self.optimizer
        x = optimizer.xr_curve.x
        a, b = optimizer.separate_params[3]
        uv_x = a*x+b
        uv_y = optimizer.uv_curve.spline(uv_x)
        return uv_y * conc_factor

class PdataProxy:
    def __init__(self, vp_analysis, treat, devel=True):
        if devel:
            from importlib import reload
            import V2PropOptimizer.V1RangeCoverter
            reload(V2PropOptimizer.V1RangeCoverter)
        from .V1RangeCoverter import convert_to_v1_model_ranges

        prop_optimizer = vp_analysis.prop_optimizer
        peaks = vp_analysis.get_current_peaks()
        paired_ranges = vp_analysis.paired_ranges
        self.cnv_ranges, ret_ranges = convert_to_v1_model_ranges(prop_optimizer, peaks, paired_ranges)
        self.paired_ranges = ret_ranges     # see convert_to_v1_model_ranges, DecompEditorFrame.make_range_info
        self.num_ranges = np.sum([len(prange.get_fromto_list()) for prange in self.cnv_ranges])
        optimizer = vp_analysis.optimizer
        self.judge_holder = JudgeHolderProxy(optimizer)
        self.is_for_sec = True
        self.conc = None
        self.conc_curves = None
        # self.mapper = create_mapper(None, treat.trimmed_sd, treat.sd_orig, treat.pre_recog)
        self.mapper = MapperProxy(optimizer)

        sd = treat.corrected_sd
        """
        consider moving these to SecSaxs.DataSet
        """
        sd.pre_recog = treat.pre_recog     # to fix AttributeError: 'DataSet' object has no attribute 'pre_recog', but not sure whether this fix is ok?
        self.conc_factor = compute_conc_factor_util()
        sd.mc_vector = self.mapper.get_conc_vector(self.conc_factor)
        sd.xr_j0 = sd.get_xr_curve().x[0]
        sd.get_xray_elution_vector()        # to set sd.xr_index

        self.sd = sd

    def make_conc_vector(self):
        # y_ = self.mapper.make_uniformly_scaled_vector(scale=1)
        self.mc_vector = self.sd.mc_vector
        return self.mc_vector

class V1PreviewAdapter:
    def __init__(self, dialog):
        self.dialog = dialog

    def show_zx_preview_using_pool(self, pdata, popts, devel=True):
        """
        adaption to PreviewButtonFrame.show_zx_preview_using_pool
        """

        if devel:
            from importlib import reload
            import LRF.LrfResultPool
            reload(LRF.LrfResultPool)       
        from LRF.LrfResultPool import LrfResultPool
        self.pool = LrfResultPool(pdata, popts)
        self.pool.run_solver(self.dialog)
        self.pool.show_dialog(self.dialog.parent)

def observe_it_as_v1_preview(vp_analysis):
    print("observe_it_as_v1_preview")

    dialog = ZxDialogProxy(vp_analysis.parent)
    adapter = V1PreviewAdapter(dialog)

    optinit_info = vp_analysis.js_canvas.optinit_info       # InitialInfo
    treat = optinit_info.treat

    pdata = PdataProxy(vp_analysis, treat)
    conc_dependence = 1
    set_setting('conc_dependence', conc_dependence)
    popts = PreviewOptions(conc_depend=conc_dependence, use_elution_models=True)
    concentration_datatype = 0      # 0: XR model, 1: XR data, 2: UV model, 3: UV data
    set_setting('concentration_datatype', concentration_datatype)

    set_setting('uv_baseline_type', 1)      # required in MappingParams.get_uv_correction_str()
    set_setting('xray_baseline_type', 1)    # 

    adapter.show_zx_preview_using_pool(pdata, popts)



# coding: utf-8
"""

    OurScreenShot.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import sys
import os
from time               import strftime
import pyautogui
from molass_legacy.KekLib.BasicUtils         import mkdirs_with_retry, auto_numbered_file, get_home_folder, get_caller_module
from molass_legacy.KekLib.TkUtils            import split_geometry
from DevSettings        import get_dev_setting

def screenshot( file=None, widget=None, log=False ):
    if file is None:
        folder = get_dev_setting( 'screenshot_folder' )
        if folder is None:
            folder = get_home_folder() + '/img'
        file = folder + '/screen-000.png'

    folder, _ = os.path.split( file  )
    if not os.path.exists( folder ):
        mkdirs_with_retry( folder )

    region = None
    if widget is not None:
        w, h, x, y = split_geometry( widget.geometry() )
        region = ( x, y, w, h )

    file = auto_numbered_file( file )
    # pyautogui.screenshot( file, region=region )

    pyautogui.screenshot( file )

    if log:
        _, f = os.path.split( file  )

        log = open( folder + '/screenshot.log', 'a' )
        _, modname = os.path.split( get_caller_module( level=2 ).__name__ )
        log.write( strftime( '%H:%M:%S ' ) + modname + ' region=' + str(region) + ' file=' + f + '\n' )
        log.close()



# coding: utf-8
"""

    ファイル名：   ChangeableLogger.py

    処理内容：

       cf.  How to change filehandle with Python logging on the fly with different classes and imports
            http://stackoverflow.com/questions/13839554/how-to-change-filehandle-with-python-logging-on-the-fly-with-different-classes-a

    Copyright (c) 2015-2020, Masatsuyo Takahashi, KEK-PF

"""
import os
import re
import logging
from io import StringIO

def arg_join( *args ):
    return ' '.join( [ str(x) for x in args ] )

class Logger:
    def __init__( self, path=None ):
        # ログの設定を行う。
        format_csv_ = '%(asctime)s,%(levelname)s,%(name)s,%(message)s'
        format_ssv_ = '%(asctime)s %(message)s'
        datefmt_    = '%Y-%m-%d %H:%M:%S'
        # logging.basicConfig( filename=path, format=format_csv_, datefmt=datefmt_ )

        if path is None:
            self.stream = StringIO()
            self.fileh = logging.StreamHandler(self.stream)
        else:
            self.stream = None
            self.fileh = logging.FileHandler( path, 'a' )
        self.formatter_csv_ = logging.Formatter( format_csv_, datefmt_ )
        self.fileh.setFormatter( self.formatter_csv_ )

        self.logger = logging.getLogger()
        self.logger.setLevel( logging.INFO )
        self.logger.addHandler( self.fileh )

        # コンソールへのログを追加する。
        self.ch = logging.StreamHandler()
        self.ch.setLevel( logging.DEBUG )
        self.formatter_ssv_ = logging.Formatter( format_ssv_, datefmt_ )
        self.ch.setFormatter( self.formatter_ssv_ )
        self.logger.addHandler( self.ch )

    def get_final_log_path( self ):
        this_dir = os.path.dirname( os.path.abspath( __file__ ) )
        log_path = os.path.abspath( this_dir + '/../../log/final_error.log' )
        n = 0
        while os.path.exists(log_path):
            n += 1
            log_path = re.sub(r'(-\d+)?\.log$', '-%02d.log'%n, log_path)
        return log_path

    def __del__( self ):
        # print( 'Logger.__del__' )
        if self.stream is not None:
            self.moveto( self.get_final_log_path() )
        self.logger.removeHandler( self.ch )
        self.logger.removeHandler( self.fileh )

    def changeto( self, path ):
        self.logger.removeHandler( self.fileh )
        self.fileh = logging.FileHandler( path, 'a' )
        self.fileh.setFormatter( self.formatter_csv_ )
        self.logger.addHandler( self.fileh )

    def moveto( self, path ):
        assert self.stream is not None
        fh = open(path, 'a')
        fh.write( self.stream.getvalue() )
        fh.close()
        self.stream = None

    def addHandler( self, handler ):
        self.logger.addHandler( handler )

    def removeHandler( self, handler ):
        self.logger.removeHandler( handler )

    def info( self, *args ):
        self.logger.info( *args )

    def warning( self, *args ):
        self.logger.warning( *args )

    def error( self, *args ):
        self.logger.error( *args )

    def debug( self, *args ):
        self.logger.debug( *args )

    def get_stream_buffer(self):
        return self.stream.getvalue()



"""
    SecTheory.EdmDemo.py

    EDM - Equilibrium Dispersive Model - Demo

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import sys
import os

# execute this file with python this-path demo
if len(sys.argv) > 1 and sys.argv[1].find("demo") >= 0:
    this_dir = os.path.dirname(os.path.abspath( __file__ ))
    sys.path.append(this_dir + '/..')

import numpy as np
import io
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Rectangle
import seaborn
seaborn.set()

from SecTheory.Edm import Edm

def demo_2d(title, a=1.5, b=0.5, Dz=0.02):
    t = np.linspace(2, 400, 400)
    edm = Edm(z=30, L=30, a=a, b=b, u=1.2, Dz=Dz)

    y = edm(t)
    fig, ax = plt.subplots()
    ax.set_title(title)
    ax.plot(t, y)
    fig.tight_layout()
    plt.show()

def peak_tailing_demo():
    a = 5
    b = 1.0
    Dz = 0.5
    demo_2d("Peak Tailing Simulation using EDM; a=%g, b=%g, Dz=%g" % (a, b, Dz), a=a, b=b, Dz=Dz)

def peak_fronting_demo():
    a = 5
    b = -0.5
    Dz = 0.05
    demo_2d("Peak Fronting Simulation using EDM; a=%g, b=%g, Dz=%g" % (a, b, Dz), a=a, b=b, Dz=Dz)

def demo_3d(title, a=1.5, b=0.5, Dz=0.02):
    t = np.linspace(2, 400, 100)
    z = np.linspace(0.5, 30, 30)
    zz, tt = np.meshgrid(z, t)
    edm = Edm(z=zz, L=30, a=a, b=b, u=1.2, Dz=Dz)
    cc = edm(tt)

    fig, ax = plt.subplots(subplot_kw=dict(projection="3d"))
    ax.set_title(title)
    ax.set_xlabel("z")
    ax.set_ylabel("t")
    ax.set_zlabel("c")
    ax.plot_surface(zz, tt, cc)
    fig.tight_layout()
    plt.show()

def peak_tailing_demo_3d():
    a = 5
    b = 1.0
    Dz = 0.5
    demo_3d("Peak Tailing Simulation using EDM; a=%g, b=%g, Dz=%g" % (a, b, Dz), a=a, b=b, Dz=Dz)

def peak_fronting_demo_3d():
    a = 5
    b = -0.5
    Dz = 0.05
    demo_3d("Peak Fronting Simulation using EDM; a=%g, b=%g, Dz=%g" % (a, b, Dz), a=a, b=b, Dz=Dz)

def animation(title, a=1.5, b=0.5, Dz=0.02, save=False):
    import matplotlib.image as mpimg
    import matplotlib.animation as animation

    P = 2000
    N = 120
    M = 90
    L = 30
    t = np.linspace(2, 400, N)

    edm = Edm(z=L, L=L, a=a, b=b, u=1.2, Dz=Dz)
    cmax = np.max(edm(t))

    z = np.linspace(1, L, M)
    zz, tt = np.meshgrid(z, t)
    edm = Edm(z=zz, L=L, a=a, b=b, u=1.2, Dz=Dz)
    cc = edm(tt)

    print("cc.shape=", cc.shape)

    fig = plt.figure(figsize=(10,5))
    fig.suptitle(title, fontsize=20)
    fig.tight_layout()
    fig.subplots_adjust(top=0.8, wspace=1.0)

    gs = GridSpec(1,4)

    ax1 = fig.add_subplot(gs[0,0])
    ax2 = fig.add_subplot(gs[0,1:4])

    ax2.set_title("Concentration at the End of Column", fontsize=16)
    ax2.set_xlim(0, t[-1])
    ax2.set_ylim(-0.01, cmax*1.2)

    images = []
    el_x = []
    el_y = []
    line, = ax2.plot(el_x, el_y, lw=5)

    LM_ratio = L/M

    for i in range(N):
        ax1.cla()
        ax1.set_title("Column State", fontsize=16)
        ax1.set_xlim(0, 1)
        ax1.set_ylim(L, 0)
        ax1.set_xticks([])

        xmin, xmax = ax1.get_xlim()
        ymin, ymax = ax1.get_ylim()

        c = cc[i,:]
        """
        r = c/np.sum(c)     # this is not appropriate for very small np.sum(c)
        n = int(r*P)
        """

        ey = 0
        for k, c in enumerate(c):
            n_ = int(P*c*LM_ratio)
            if k == M - 1:
                ey += n_
            if n_ > 0:
                pos = np.random.uniform(0, 1, (n_,2))
                x_ = pos[:,0]
                y_ = (k + pos[:,1])*LM_ratio
                ax1.plot(x_, y_, "o", color="red", markersize=1)

        el_x.append(t[i])
        el_y.append(ey/P/LM_ratio)
        line.set_xdata(el_x)
        line.set_ydata(el_y)

        plt.pause(0.2)

        buffer = io.BytesIO()
        fig.savefig(buffer)
        buffer.seek(0)
        img = mpimg.imread(buffer)
        images.append(img)

    plt.show()
    plt.close()

    if not save:
        return

    print("saving the animation ...")

    fig, ax = plt.subplots(figsize=(9,5))
    ax.set_axis_off()
    fig.tight_layout()

    ims = []
    for img in images:
        print(".")
        im = ax.imshow(img, animated=True)
        ims.append([im])

    ani = animation.ArtistAnimation(fig, ims, interval=100, blit=True, repeat_delay=1000)
    ani.save("animation.gif")

    plt.show()

def peak_taling_animation(save=False):
    animation("Peak Tailing Simulation using EDM", a=5, b=1.0, Dz=0.5, save=save)

def peak_fronting_animation(save=False):
    animation("Peak Fronting Simulation using EDM", a=5, b=-0.5, Dz=0.05, save=save)

def paper_proof_1():
    t = np.linspace(0, 40, 100)
    z = 10
    L = 10
    u = 1.2
    Pe = 600
    Dz = L*u/Pe
    print("Dz=", Dz)

    fig, ax = plt.subplots()

    for b in [0.5, 0.2, 1.0, 2.0]:
        edm = Edm(z=z, b=b, Dz=Dz)
        c = edm(t)
        ax.plot(t, c, label="b=%.1f" % b)

    ax.legend()
    fig.tight_layout()
    plt.show()

def emg_vs_edm():
    from scipy.optimize import minimize

    for name in ["KekLib", "DataStructure", "SerialAnalyzer", "Decomposer"]:
        sys.path.append(this_dir + '/../' + name)

    from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks
    from molass_legacy.QuickAnalysis.ModeledPeaks import egh

    title = "EGH vs. EDM"
    a = 5
    b = -0.5
    Dz = 0.05
    x = np.linspace(2, 400, 400)
    edm = Edm(z=30, L=30, a=a, b=b, u=1.2, Dz=Dz)

    y = edm(x)

    fig, axes = plt.subplots(ncols=3, figsize=(18,5))
    fig.suptitle(title, fontsize=20)

    num_peaks = 1

    def objective(p):
        peaks = p.reshape(num_peaks, 4)
        ty_ = np.zeros(len(x))
        for h, m, s, t in peaks:
            y_ = egh(x, h, m, s, t)
            ty_ += y_
        return np.sum((ty_ - y)**2)

    for n, ax in enumerate(axes):
        num_peaks = n + 1
        ax.set_title("Number of components = %d" % num_peaks, fontsize=16)

        peaks = recognize_peaks(x, y, exact_num_peaks=num_peaks)

        ax.plot(x, y, color="orange", label="EDM")

        params = np.array(peaks)

        ret = minimize(objective, params.flatten())
        peaks = ret.x.reshape(num_peaks, 4)

        ty_ = np.zeros(len(x))
        for k, (h, m, s, t) in enumerate(peaks):
            y_ = egh(x, h, m, s, t)
            ty_ += y_
            ax.plot(x, y_, ":", label="EGH-%d" % (k+1))

        ax.plot(x, ty_, ":", color="red", lw=2, label="EGH total")
        ax.legend(loc="upper left")

        xmin, xmax = ax.get_xlim()
        ylim = ax.get_ylim()

        rmsd = np.sqrt(np.mean((ty_ - y)))
        tx = xmin*0.7 + xmax*0.3
        ty = np.mean(ylim)
        ax.text(tx, ty, "RMSD=%.2g" % rmsd, alpha=0.5, fontsize=20, ha="center", va="center")

    fig.tight_layout()
    plt.show()

def quadratic_isotherm():

    def q(a, b, c):
        return (a*c)/(1 + b*c)

    def q_(a, b, c, c0):
        denominator = (1 + b*c0)**3
        gam1 = (a * b**2 * c0**3) / denominator
        gam2 = (a + 3*a*b*c0) / denominator
        gam3 = (-a*b) / denominator
        return gam1 + gam2*c + gam3*c**2

    a = 1
    b = 0.5
    c0 = 0.0001
    c = np.linspace(0, 0.6, 100)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
    fig.suptitle("Quadratic isotherm")
    ymin, ymax = 1, 0
    for ax, a_, b_, c0_ in [(ax1, a, b, c0), (ax2, a, -b, c0)]:
        ax.set_title("a, b, c0 = (%g, %g, %g)" % (a_, b_, c0_))
        ax.set_xlabel("mobile phase concentration")
        ax.set_ylabel("adsorbed phase concentration")
        ax.plot(c, q(a_, b_, c), label="Langmuir isotherm")
        ax.plot(c, q_(a_, b_, c, c0_), label="Quadratic isotherm")
        ymin_, ymax_ = ax.get_ylim()
        ymin = min(ymin, ymin_)
        ymax = max(ymax, ymax_)
        ax.legend(loc="upper left")

    for ax in ax1, ax2:
        ax.set_ylim(ymin, ymax)

    fig.tight_layout()
    plt.show()

if __name__ == '__main__':
    # peak_tailing_demo()
    # peak_tailing_demo_3d()
    # peak_fronting_demo()
    peak_fronting_demo_3d()
    # peak_taling_animation(save=True)
    # peak_fronting_animation(save=True)
    # paper_proof_1()
    # emg_vs_edm()
    # quadratic_isotherm()
